# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import textwrap
from random import Random

import pytest

from hypothesis import HealthCheck, assume, given, settings
from hypothesis.errors import Flaky
from hypothesis.internal.conjecture.choice import ChoiceTemplate
from hypothesis.internal.conjecture.data import ConjectureData, Status, StopTest
from hypothesis.internal.conjecture.datatree import (
    Branch,
    DataTree,
    compute_max_children,
)
from hypothesis.internal.conjecture.engine import ConjectureRunner
from hypothesis.internal.conjecture.floats import float_to_int
from hypothesis.internal.conjecture.provider_conformance import (
    boolean_constraints,
    integer_constraints,
)
from hypothesis.internal.floats import next_up
from hypothesis.vendor import pretty

from tests.conjecture.common import (
    constraints_strategy,
    fresh_data,
    interesting_origin,
    nodes,
    run_to_nodes,
)

runner_settings = settings(
    max_examples=100, database=None, suppress_health_check=list(HealthCheck)
)


def runner_for(*examples):
    def accept(tf):
        runner = ConjectureRunner(
            tf, settings=runner_settings, random=Random(0), ignore_limits=True
        )
        ran_examples = []
        for choices in examples:
            data = runner.cached_test_function(choices)
            ran_examples.append((choices, data))
        for choices, d in ran_examples:
            rewritten, status = runner.tree.rewrite(choices)
            assert status == d.status
            assert rewritten == d.choices
        return runner

    return accept


def test_can_lookup_cached_examples():
    @runner_for((0, 0), (0, 1))
    def runner(data):
        data.draw_integer()
        data.draw_integer()


def test_can_lookup_cached_examples_with_forced():
    @runner_for((0, 0), (0, 1))
    def runner(data):
        data.draw_integer(forced=1)
        data.draw_integer()


def test_can_detect_when_tree_is_exhausted():
    @runner_for((False,), (True,))
    def runner(data):
        data.draw_boolean()

    assert runner.tree.is_exhausted


def test_can_detect_when_tree_is_exhausted_variable_size():
    @runner_for((False,), (True, False), (True, True))
    def runner(data):
        if data.draw_boolean():
            data.draw_boolean()

    assert runner.tree.is_exhausted


def test_one_dead_branch():
    @runner_for(*([(0, i) for i in range(16)] + [(i,) for i in range(1, 16)]))
    def runner(data):
        i = data.draw_integer(0, 15)
        if i > 0:
            data.mark_invalid()
        data.draw_integer(0, 15)

    assert runner.tree.is_exhausted


def test_non_dead_root():
    @runner_for((False, False), (True, False), (True, True))
    def runner(data):
        data.draw_boolean()
        data.draw_boolean()


def test_can_reexecute_dead_examples():
    @runner_for((False, False), (False, True), (False, False))
    def runner(data):
        data.draw_boolean()
        data.draw_boolean()


def test_novel_prefixes_are_novel():
    def tf(data):
        for _ in range(4):
            data.draw_bytes(1, 1, forced=b"\0")
            data.draw_integer(0, 3)

    runner = ConjectureRunner(
        tf, settings=settings(runner_settings, max_examples=1000), random=Random(0)
    )
    for _ in range(100):
        prefix = runner.tree.generate_novel_prefix(runner.random)
        extension = prefix + (ChoiceTemplate("simplest", count=100),)
        assert runner.tree.rewrite(extension)[1] is None
        result = runner.cached_test_function(extension)
        assert runner.tree.rewrite(extension)[0] == result.choices


def test_overruns_if_prefix():
    runner = ConjectureRunner(
        lambda data: [data.draw_boolean() for _ in range(2)],
        settings=runner_settings,
        random=Random(0),
    )
    runner.cached_test_function([False, False])
    assert runner.tree.rewrite([False])[1] is Status.OVERRUN


def test_stores_the_tree_flat_until_needed():
    @runner_for((False,) * 10)
    def runner(data):
        for _ in range(10):
            data.draw_boolean()
        data.mark_interesting(interesting_origin())

    root = runner.tree.root
    assert len(root.constraints) == 10
    assert len(root.values) == 10
    assert root.transition.status == Status.INTERESTING


def test_split_in_the_middle():
    @runner_for((0, 0, 2), (0, 1, 3))
    def runner(data):
        data.draw_integer(0, 1)
        data.draw_integer(0, 1)
        data.draw_integer(0, 15)
        data.mark_interesting(interesting_origin())

    root = runner.tree.root
    assert len(root.constraints) == len(root.values) == 1
    assert list(root.transition.children[0].values) == [2]
    assert list(root.transition.children[1].values) == [3]


def test_stores_forced_nodes():
    @runner_for((0, 0, 0))
    def runner(data):
        data.draw_integer(0, 1, forced=0)
        data.draw_integer(0, 1)
        data.draw_integer(0, 1, forced=0)
        data.mark_interesting(interesting_origin())

    root = runner.tree.root
    assert root.forced == {0, 2}


def test_correctly_relocates_forced_nodes():
    @runner_for((0, 0), (1, 0))
    def runner(data):
        data.draw_integer(0, 1)
        data.draw_integer(0, 1, forced=0)
        data.mark_interesting(interesting_origin())

    root = runner.tree.root
    assert root.transition.children[1].forced == {0}
    assert root.transition.children[0].forced == {0}


def test_can_go_from_interesting_to_valid():
    tree = DataTree()
    data = ConjectureData.for_choices([], observer=tree.new_observer())
    with pytest.raises(StopTest):
        data.conclude_test(Status.INTERESTING)

    data = ConjectureData.for_choices([], observer=tree.new_observer())
    with pytest.raises(StopTest):
        data.conclude_test(Status.VALID)


def test_going_from_interesting_to_invalid_is_flaky():
    tree = DataTree()
    data = ConjectureData.for_choices([], observer=tree.new_observer())
    with pytest.raises(StopTest):
        data.conclude_test(Status.INTERESTING)

    data = ConjectureData.for_choices([], observer=tree.new_observer())
    with pytest.raises(Flaky):
        data.conclude_test(Status.INVALID)


def test_concluding_at_prefix_is_flaky():
    tree = DataTree()
    data = ConjectureData.for_choices((True,), observer=tree.new_observer())
    data.draw_boolean()
    with pytest.raises(StopTest):
        data.conclude_test(Status.INTERESTING)

    data = ConjectureData.for_choices([], observer=tree.new_observer())
    with pytest.raises(Flaky):
        data.conclude_test(Status.INVALID)


def test_concluding_with_overrun_at_prefix_is_not_flaky():
    tree = DataTree()
    data = ConjectureData.for_choices((True,), observer=tree.new_observer())
    data.draw_boolean()
    with pytest.raises(StopTest):
        data.conclude_test(Status.INTERESTING)

    data = ConjectureData.for_choices([], observer=tree.new_observer())
    with pytest.raises(StopTest):
        data.conclude_test(Status.OVERRUN)


def test_changing_n_bits_is_flaky_in_prefix():
    tree = DataTree()
    data = ConjectureData.for_choices((1,), observer=tree.new_observer())
    data.draw_integer(0, 1)
    with pytest.raises(StopTest):
        data.conclude_test(Status.INTERESTING)

    data = ConjectureData.for_choices((1,), observer=tree.new_observer())
    with pytest.raises(Flaky):
        data.draw_integer(0, 3)


def test_changing_n_bits_is_flaky_in_branch():
    tree = DataTree()

    for i in [0, 1]:
        data = ConjectureData.for_choices((i,), observer=tree.new_observer())
        data.draw_integer(0, 1)
        with pytest.raises(StopTest):
            data.conclude_test(Status.INTERESTING)

    data = ConjectureData.for_choices((1,), observer=tree.new_observer())
    with pytest.raises(Flaky):
        data.draw_integer(0, 3)


def test_extending_past_conclusion_is_flaky():
    tree = DataTree()
    data = ConjectureData.for_choices((True,), observer=tree.new_observer())
    data.draw_boolean()
    with pytest.raises(StopTest):
        data.conclude_test(Status.INTERESTING)

    data = ConjectureData.for_choices((True, False), observer=tree.new_observer())
    data.draw_boolean()

    with pytest.raises(Flaky):
        data.draw_boolean()


def test_changing_to_forced_is_flaky():
    tree = DataTree()
    data = ConjectureData.for_choices((True,), observer=tree.new_observer())
    data.draw_boolean()
    with pytest.raises(StopTest):
        data.conclude_test(Status.INTERESTING)

    data = ConjectureData.for_choices((True, False), observer=tree.new_observer())

    with pytest.raises(Flaky):
        data.draw_boolean(forced=True)


def test_changing_value_of_forced_is_flaky():
    tree = DataTree()
    data = ConjectureData.for_choices((True,), observer=tree.new_observer())
    data.draw_boolean(forced=True)
    with pytest.raises(StopTest):
        data.conclude_test(Status.INTERESTING)

    data = ConjectureData.for_choices((True, False), observer=tree.new_observer())

    with pytest.raises(Flaky):
        data.draw_boolean(forced=False)


def test_does_not_truncate_if_unseen():
    tree = DataTree()
    nodes = (1, 2, 3, 4)
    assert tree.rewrite(nodes) == (nodes, None)


def test_truncates_if_seen():
    tree = DataTree()

    nodes = (1, 2, 3, 4)

    data = ConjectureData.for_choices(nodes, observer=tree.new_observer())
    data.draw_integer()
    data.draw_integer()
    data.freeze()

    assert tree.rewrite(nodes) == (nodes[:2], Status.VALID)


def test_child_becomes_exhausted_after_split():
    tree = DataTree()
    data = ConjectureData.for_choices((b"\0", b"\0"), observer=tree.new_observer())
    data.draw_bytes(1, 1)
    data.draw_bytes(1, 1, forced=b"\0")
    data.freeze()

    data = ConjectureData.for_choices((b"\1", b"\0"), observer=tree.new_observer())
    data.draw_bytes(1, 1)
    data.draw_bytes(1, 1)
    data.freeze()

    assert not tree.is_exhausted
    assert tree.root.transition.children[b"\0"].is_exhausted


def test_will_generate_novel_prefix_to_avoid_exhausted_branches():
    tree = DataTree()
    data = ConjectureData.for_choices((1,), observer=tree.new_observer())
    data.draw_integer(0, 1)
    data.freeze()

    data = ConjectureData.for_choices((0, b"\1"), observer=tree.new_observer())
    data.draw_integer(0, 1)
    data.draw_bytes(1, 1)
    data.freeze()

    prefix = tree.generate_novel_prefix(Random(0))

    assert len(prefix) == 2
    assert prefix[0] == 0


def test_will_mark_changes_in_discard_as_flaky():
    tree = DataTree()
    data = ConjectureData.for_choices((1, 1), observer=tree.new_observer())
    data.start_span(10)
    data.draw_integer(0, 1)
    data.stop_span()
    data.draw_integer(0, 1)
    data.freeze()

    data = ConjectureData.for_choices((1, 1), observer=tree.new_observer())
    data.start_span(10)
    data.draw_integer(0, 1)

    with pytest.raises(Flaky):
        data.stop_span(discard=True)


def test_is_not_flaky_on_positive_zero_and_negative_zero():
    # if we store floats in a naive way, the 0.0 and -0.0 draws will be treated
    # equivalently and will lead to flaky errors when they diverge on the boolean
    # draw.
    tree = DataTree()

    data = ConjectureData.for_choices((0.0, False), observer=tree.new_observer())
    f = data.draw_float()
    assert float_to_int(f) == float_to_int(0.0)
    data.draw_boolean()
    data.freeze()

    data = ConjectureData.for_choices((-0.0, True), observer=tree.new_observer())
    f = data.draw_float()
    assert float_to_int(f) == float_to_int(-0.0)
    data.draw_boolean()
    data.freeze()

    assert isinstance(tree.root.transition, Branch)
    children = tree.root.transition.children
    assert children[float_to_int(0.0)].values == [False]
    assert children[float_to_int(-0.0)].values == [True]


def test_low_probabilities_are_still_explored():
    tree = DataTree()
    data = ConjectureData.for_choices([False], observer=tree.new_observer())
    data.draw_boolean(p=1e-10)  # False

    prefix = tree.generate_novel_prefix(Random())
    assert prefix[0]


def _test_observed_draws_are_recorded_in_tree(choice_type):
    @given(constraints_strategy(choice_type))
    def test(constraints):
        # we currently split pseudo-choices with a single child into their
        # own transition, which clashes with our asserts below. If we ever
        # change this (say, by not writing pseudo choices to the ir at all),
        # this restriction can be relaxed.
        assume(compute_max_children(choice_type, constraints) > 1)

        tree = DataTree()
        data = fresh_data(observer=tree.new_observer())
        draw_func = getattr(data, f"draw_{choice_type}")
        draw_func(**constraints)

        assert tree.root.transition is None
        assert tree.root.choice_types == [choice_type]

    test()


def _test_non_observed_draws_are_not_recorded_in_tree(choice_type):
    @given(constraints_strategy(choice_type))
    def test(constraints):
        assume(compute_max_children(choice_type, constraints) > 1)

        tree = DataTree()
        data = fresh_data(observer=tree.new_observer())
        draw_func = getattr(data, f"draw_{choice_type}")
        draw_func(**constraints, observe=False)

        root = tree.root
        assert root.transition is None
        assert root.constraints == root.values == root.choice_types == []

    test()


@pytest.mark.parametrize(
    "choice_type", ["integer", "float", "boolean", "string", "bytes"]
)
def test_observed_choice_type_draw(choice_type):
    _test_observed_draws_are_recorded_in_tree(choice_type)


@pytest.mark.parametrize(
    "choice_type", ["integer", "float", "boolean", "string", "bytes"]
)
def test_non_observed_choice_type_draw(choice_type):
    _test_non_observed_draws_are_not_recorded_in_tree(choice_type)


def test_can_generate_hard_values():
    tree = DataTree()

    min_value = 0
    max_value = 1000
    # set up `tree` such that [0, 999] have been drawn and only n=1000 remains.
    for i in range(max_value):
        data = ConjectureData.for_choices([i], observer=tree.new_observer())
        data.draw_integer(min_value, max_value)
        data.freeze()

    # this test doubles as conjecture coverage for using our child cache, so
    # ensure we don't miss that logic by getting lucky and drawing the correct
    # value once or twice.
    for _ in range(20):
        prefix = tree.generate_novel_prefix(Random())
        data = ConjectureData.for_choices(prefix)
        assert data.draw_integer(min_value, max_value) == 1000


def test_can_generate_hard_floats():
    # similar to test_can_generate_hard_values, but exercises float-specific
    # logic for handling e.g. 0.0 vs -0.0 as different keys.
    tree = DataTree()

    def next_up_n(f, n):
        for _ in range(n):
            f = next_up(f)
        return f

    min_value = -0.0
    max_value = next_up_n(min_value, 100)
    for n in range(100):

        @run_to_nodes
        def nodes(data):
            f = next_up_n(min_value, n)
            data.draw_float(min_value, max_value, forced=f, allow_nan=False)
            data.mark_interesting(interesting_origin())

        data = ConjectureData.for_choices(
            [n.value for n in nodes], observer=tree.new_observer()
        )
        data.draw_float(min_value, max_value, allow_nan=False)
        data.freeze()

    # we have left out a single value, so we can assert that generate_novel_prefix
    # is equal to that value.
    #
    # this test doubles as conjecture coverage for drawing floats from the
    # children cache. Draw a few times to ensure we hit that logic (as opposed
    # to getting lucky and drawing the correct value the first time).
    for _ in range(20):
        expected_value = next_up_n(min_value, 100)
        prefix = tree.generate_novel_prefix(Random())
        data = ConjectureData.for_choices(prefix)
        assert data.draw_float(min_value, max_value, allow_nan=False) == expected_value


@given(boolean_constraints(), integer_constraints())
def test_datatree_repr(bool_constraints, int_constraints):
    tree = DataTree()

    origin = interesting_origin()

    observer = tree.new_observer()
    observer.draw_boolean(True, was_forced=False, constraints=bool_constraints)
    observer.conclude_test(Status.INVALID, interesting_origin=None)

    observer = tree.new_observer()
    observer.draw_boolean(False, was_forced=False, constraints=bool_constraints)
    observer.draw_integer(42, was_forced=False, constraints=int_constraints)
    observer.conclude_test(Status.VALID, interesting_origin=None)

    observer = tree.new_observer()
    observer.draw_boolean(False, was_forced=False, constraints=bool_constraints)
    observer.draw_integer(0, was_forced=False, constraints=int_constraints)
    observer.draw_boolean(False, was_forced=True, constraints=bool_constraints)
    observer.conclude_test(Status.INTERESTING, interesting_origin=origin)

    observer = tree.new_observer()
    observer.draw_boolean(False, was_forced=False, constraints=bool_constraints)
    observer.draw_integer(5, was_forced=False, constraints=int_constraints)

    assert (
        pretty.pretty(tree)
        == textwrap.dedent(
            f"""
        boolean True {bool_constraints}
          Conclusion (Status.INVALID)
        boolean False {bool_constraints}
          integer 42 {int_constraints}
            Conclusion (Status.VALID)
          integer 0 {int_constraints}
            boolean False [forced] {bool_constraints}
              Conclusion (Status.INTERESTING, {origin})
          integer 5 {int_constraints}
            unknown
        """
        ).strip()
    )


def _draw(data, node, *, forced=None):
    return getattr(data, f"draw_{node.type}")(**node.constraints, forced=forced)


@given(nodes(was_forced=True, choice_types=["float"]))
def test_simulate_forced_floats(node):
    tree = DataTree()

    data = ConjectureData.for_choices([node.value], observer=tree.new_observer())
    data.draw_float(**node.constraints, forced=node.value)
    with pytest.raises(StopTest):
        data.conclude_test(Status.VALID)

    data = ConjectureData.for_choices([node.value], observer=tree.new_observer())
    tree.simulate_test_function(data)
    data.freeze()
    assert data.nodes == (node,)



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import subprocess

BEFORE = """
from hypothesis.strategies import complex_numbers, complex_numbers as cn

complex_numbers(min_magnitude=None)  # simple call to fix
complex_numbers(min_magnitude=None, max_magnitude=1)  # plus arg after
complex_numbers(allow_infinity=False, min_magnitude=None)  # plus arg before
cn(min_magnitude=None)  # imported as alias
"""
AFTER = BEFORE.replace("None", "0")
_unchanged = """
complex_numbers(min_magnitude=1)  # value OK

class Foo:
    def complex_numbers(self, **kw): pass

    complex_numbers(min_magnitude=None)  # defined in a different scope
"""
BEFORE += _unchanged
AFTER += _unchanged
del _unchanged


def run(command, *, cwd=None, input=None):
    return subprocess.run(
        command,
        input=input,
        capture_output=True,
        shell=True,
        text=True,
        cwd=cwd,
        encoding="utf-8",
    )


def test_codemod_single_file(tmp_path):
    fname = tmp_path / "mycode.py"
    fname.write_text(BEFORE, encoding="utf-8")
    result = run("hypothesis codemod mycode.py", cwd=tmp_path)
    assert result.returncode == 0
    assert fname.read_text(encoding="utf-8") == AFTER


def test_codemod_multiple_files(tmp_path):
    # LibCST had some trouble with multiprocessing on Windows
    files = [tmp_path / "mycode1.py", tmp_path / "mycode2.py"]
    for f in files:
        f.write_text(BEFORE, encoding="utf-8")
    result = run("hypothesis codemod mycode1.py mycode2.py", cwd=tmp_path)
    assert result.returncode == 0
    for f in files:
        assert f.read_text(encoding="utf-8") == AFTER


def test_codemod_from_stdin():
    result = run("hypothesis codemod -", input=BEFORE)
    assert result.returncode == 0
    assert result.stdout.rstrip() == AFTER.rstrip()



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import re
from pathlib import Path

import pytest

from hypothesistooling.__main__ import PYTHONS

ci_checks = "    ".join(
    line.strip()
    for line in Path(".github/workflows/main.yml")
    .read_text(encoding="utf-8")
    .splitlines()
    if "- check-py" in line
)


@pytest.mark.parametrize("version", sorted(PYTHONS))
def test_python_versions_are_tested_in_ci(version):
    slug = version.replace("pypy", "py").replace(".", "")
    print(ci_checks)
    assert f"- check-py{slug}" in ci_checks, f"Add {version} to main.yml and tox.ini"


def test_python_versions_are_in_trove_classifiers():
    got_classifiers = {
        line.strip(' ",\n')
        for line in Path("hypothesis-python/pyproject.toml")
        .read_text(encoding="utf-8")
        .splitlines()
        if "Programming Language :: Python :: 3." in line
    }
    expected_classifiers = {
        f"Programming Language :: Python :: 3.{v.split('.')[1]}"
        for v in PYTHONS.values()
        if re.fullmatch(r"3\.\d+\.\d+", v)
    }
    assert got_classifiers == expected_classifiers



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import datetime
import re
import sys
import types
from pathlib import Path

from docutils import nodes
from sphinx.util.docutils import SphinxRole

root = Path(__file__).parent.parent
# insert so the local versions are consulted first, before any installed version
sys.path.insert(0, str(root / "src"))
sys.path.insert(0, str(Path(__file__).parent / "_ext"))

needs_sphinx = re.search(
    r"sphinx==([0-9\.]+)", root.joinpath("../requirements/tools.txt").read_text()
).group(1)
default_role = "py:obj"
nitpicky = True

autodoc_member_order = "bysource"
autodoc_typehints = "none"
maximum_signature_line_length = 60  # either one line, or one param per line

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.extlinks",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx_codeautolink",
    "sphinx_selective_exclude.eager_only",
    "sphinx-jsonschema",
    # loading this extension overrides the default -b linkcheck behavior with
    # custom url ignore logic. see hypothesis_linkcheck.py for details.
    "hypothesis_linkcheck",
    "hypothesis_redirects",
]

templates_path = ["_templates"]

# config for hypothesis_redirects
redirects = {
    "details": "reference/index.html",
    "data": "reference/strategies.html",
    "database": "reference/api.html#database",
    # "stateful": "reference/api.html#stateful-tests",
    "reproducing": "reference/api.html",
    "ghostwriter": "reference/integrations.html#ghostwriter",
    "django": "reference/strategies.html#django",
    "numpy": "reference/strategies.html#numpy",
    "observability": "reference/integrations.html#observability",
    "settings": "reference/api.html#settings",
    "endorsements": "usage.html#testimonials",
    # TODO enable when we actually rename them
    # "extras": "extensions.html",
    "supported": "compatibility.html",
    "changes": "changelog.html",
    "strategies": "extensions.html",
    # these pages were removed without replacement
    "support": "index.html",
    "manifesto": "index.html",
    "examples": "index.html",
}
redirect_html_template_file = "redirect.html.template"

source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "Hypothesis"
author = "the Hypothesis team"
copyright = f"2013-{datetime.date.today().year}, {author}"

_d = {}
_version_file = root.joinpath("src", "hypothesis", "version.py")
exec(_version_file.read_text(encoding="utf-8"), _d)
version = _d["__version__"]
release = _d["__version__"]


# custom role for version syntaxes.
# :v:`6.131.0`       = [v6.131.0](changelog.html#v6.13.0)
# :version:`6.131.0` = [version 6.131.0](changelog.html#v6.13.0)
class VersionRole(SphinxRole):
    def __init__(self, prefix):
        self.prefix = prefix

    def run(self):
        node = nodes.reference(
            "",
            f"{self.prefix}{self.text}",
            refuri=f"changelog.html#v{self.text.replace('.', '-')}",
        )
        return [node], []


def setup(app):
    if root.joinpath("RELEASE.rst").is_file():
        app.tags.add("has_release_file")

    # Workaround for partial-initialization problem when autodoc imports libcst
    import libcst

    import hypothesis.extra.codemods

    assert libcst
    assert hypothesis.extra.codemods

    # patch in mock array_api namespace so we can autodoc it
    from hypothesis.extra.array_api import (
        RELEASED_VERSIONS,
        make_strategies_namespace,
        mock_xp,
    )

    mod = types.ModuleType("xps")
    mod.__dict__.update(
        make_strategies_namespace(mock_xp, api_version=RELEASED_VERSIONS[-1]).__dict__
    )
    assert "xps" not in sys.modules
    sys.modules["xps"] = mod

    def process_signature(app, what, name, obj, options, signature, return_annotation):
        # manually override an ugly signature from .. autofunction. Alternative we
        # could manually document with `.. function:: run_conformance_test(...)`,
        # but that's less likely to stay up to date.
        if (
            name
            == "hypothesis.internal.conjecture.provider_conformance.run_conformance_test"
        ):
            # so we know if this ever becomes obsolete
            assert "_realize_objects" in signature
            signature = re.sub(
                r"_realize_objects=.*",
                "_realize_objects=st.from_type(object) | st.from_type(type).flatmap(st.from_type))",
                signature,
            )
        return signature, return_annotation

    app.connect("autodoc-process-signature", process_signature)
    app.add_role("v", VersionRole(prefix="v"))
    app.add_role("version", VersionRole(prefix="version "))


language = "en"
exclude_patterns = ["_build", "prolog.rst"]
pygments_style = "sphinx"
todo_include_todos = False

# To run linkcheck (last argument is the output dir)
#   sphinx-build --builder linkcheck hypothesis-python/docs linkcheck
linkcheck_ignore = [
    # we'll assume that python isn't going to break peps, and github isn't going
    # to break issues/pulls. (and if they did, we'd hopefully notice quickly).
    r"https://peps.python.org/pep-.*",
    r"https://github.com/HypothesisWorks/hypothesis/issues/\d+",
    r"https://github.com/HypothesisWorks/hypothesis/pull/\d+",
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "pandas": ("https://pandas.pydata.org/docs/", None),
    "pytest": ("https://docs.pytest.org/en/stable/", None),
    "django": (
        "http://docs.djangoproject.com/en/stable/",
        "http://docs.djangoproject.com/en/stable/_objects/",
    ),
    "dateutil": ("https://dateutil.readthedocs.io/en/stable/", None),
    "redis": ("https://redis.readthedocs.io/en/stable/", None),
    "attrs": ("https://www.attrs.org/en/stable/", None),
    "sphinx": ("https://www.sphinx-doc.org/en/master/", None),
    "IPython": ("https://ipython.readthedocs.io/en/stable/", None),
    "lark": ("https://lark-parser.readthedocs.io/en/stable/", None),
    "xarray": ("https://docs.xarray.dev/en/stable/", None),
}

autodoc_mock_imports = ["numpy", "pandas", "redis", "django", "pytz"]

rst_prolog = (Path(__file__).parent / "prolog.rst").read_text()

codeautolink_autodoc_inject = False
codeautolink_global_preface = """
from hypothesis import *
import hypothesis.strategies as st
from hypothesis.strategies import *
"""

# This config value must be a dictionary of external sites, mapping unique
# short alias names to a base URL and a prefix.
# See http://sphinx-doc.org/ext/extlinks.html
_repo = "https://github.com/HypothesisWorks/hypothesis/"
extlinks = {
    "commit": (_repo + "commit/%s", "commit %s"),
    "gh-file": (_repo + "blob/master/%s", "%s"),
    "gh-link": (_repo + "%s", "%s"),
    "issue": (_repo + "issues/%s", "issue #%s"),
    "pull": (_repo + "pull/%s", "pull request #%s"),
    "pypi": ("https://pypi.org/project/%s/", "%s"),
    "bpo": ("https://bugs.python.org/issue%s", "bpo-%s"),
    "xp-ref": ("https://data-apis.org/array-api/latest/API_specification/%s", "%s"),
    "wikipedia": ("https://en.wikipedia.org/wiki/%s", "%s"),
}

# -- Options for HTML output ----------------------------------------------

html_theme = "furo"
# remove "Hypothesis <version> documentation" from just below logo on the sidebar
html_theme_options = {"sidebar_hide_name": True}
html_static_path = ["_static"]
html_css_files = [
    "better-signatures.css",
    "wrap-in-tables.css",
    "no-scroll.css",
    "dark-fix.css",
]
htmlhelp_basename = "Hypothesisdoc"
html_favicon = "../../brand/favicon.ico"
html_logo = "../../brand/dragonfly-rainbow-150w.svg"


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {}
latex_documents = [
    (master_doc, "Hypothesis.tex", "Hypothesis Documentation", author, "manual")
]
man_pages = [(master_doc, "hypothesis", "Hypothesis Documentation", [author], 1)]
texinfo_documents = [
    (
        master_doc,
        "Hypothesis",
        "Hypothesis Documentation",
        author,
        "Hypothesis",
        "Advanced property-based testing for Python.",
        "Miscellaneous",
    )
]



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

from hypothesis import assume, given, strategies as st
from hypothesis.strategies._internal.lazy import unwrap_strategies

from tests.common.debug import assert_no_examples


@given(st.integers().map(lambda x: assume(x % 3 != 0) and x))
def test_can_assume_in_map(x):
    assert x % 3 != 0


def test_assume_in_just_raises_immediately():
    assert_no_examples(st.just(1).map(lambda x: assume(x == 2)))


def test_identity_map_is_noop():
    s = unwrap_strategies(st.integers())
    assert s.map(lambda x: x) is s



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import pytest

from hypothesis import given, note, settings, strategies as st
from hypothesis.errors import HypothesisWarning, InvalidArgument

from tests.common.debug import (
    assert_all_examples,
    assert_no_examples,
    check_can_generate_examples,
    find_any,
    minimal,
)
from tests.common.utils import checks_deprecated_behaviour


@given(st.recursive(st.booleans(), st.lists, max_leaves=10))
def test_respects_leaf_limit(xs):
    def flatten(x):
        if isinstance(x, list):
            return sum(map(flatten, x), [])
        else:
            return [x]

    assert len(flatten(xs)) <= 10


def test_can_find_nested():
    x = minimal(
        st.recursive(st.booleans(), lambda x: st.tuples(x, x)),
        lambda x: isinstance(x, tuple) and isinstance(x[0], tuple),
    )

    assert x == ((False, False), False)


def test_recursive_call_validates_expand_returns_strategies():
    with pytest.raises(InvalidArgument):
        check_can_generate_examples(st.recursive(st.booleans(), lambda x: 1))


def test_recursive_call_validates_base_is_strategy():
    x = st.recursive(1, lambda x: st.none())
    with pytest.raises(InvalidArgument):
        check_can_generate_examples(x)


def test_can_find_exactly_max_leaves():
    strat = st.recursive(st.none(), lambda x: st.tuples(x, x), max_leaves=5)

    def enough_leaves(t):
        print(t)
        count = 0
        stack = [t]
        while stack:
            s = stack.pop()
            if s is None:
                count += 1
            else:
                stack.extend(s)
        return count >= 5

    find_any(strat, enough_leaves)


@given(st.recursive(st.none(), lambda x: st.tuples(x, x), max_leaves=1))
def test_can_exclude_branching_with_max_leaves(t):
    assert t is None


@given(st.recursive(st.none(), lambda x: st.one_of(x, x)))
def test_issue_1502_regression(s):
    pass


def test_recursive_can_generate_varied_structures():
    values = st.recursive(st.none(), st.lists)

    find_any(values, lambda x: x is None)
    find_any(values, lambda x: isinstance(x, list))
    find_any(
        values, lambda x: isinstance(x, list) and any(isinstance(y, list) for y in x)
    )


@checks_deprecated_behaviour
def test_recursive_can_generate_varied_structures_without_using_leaves():
    values = st.recursive(st.none(), lambda _: st.lists(st.none()))

    find_any(values, lambda x: x is None)
    find_any(values, lambda x: isinstance(x, list))
    # The bad `extend` function means we can't actually recurse!
    assert_no_examples(
        values, lambda x: isinstance(x, list) and any(isinstance(y, list) for y in x)
    )


@pytest.mark.parametrize(
    "s",
    [
        st.recursive(None, st.lists),
        st.recursive(st.none(), lambda x: None),
        st.recursive(st.none(), st.lists, max_leaves=-1),
        st.recursive(st.none(), st.lists, max_leaves=0),
        st.recursive(st.none(), st.lists, max_leaves=1.0),
        st.recursive(st.none(), st.lists, min_leaves=-1),
        st.recursive(st.none(), st.lists, min_leaves=0),
        st.recursive(st.none(), st.lists, min_leaves=1.0),
        st.recursive(st.none(), st.lists, min_leaves=10, max_leaves=5),
        st.recursive(st.none(), lambda _: st.lists(st.none()), min_leaves=1),
    ],
)
def test_invalid_args(s):
    with pytest.raises(InvalidArgument):
        check_can_generate_examples(s)


def _count_leaves(tree):
    if isinstance(tree, tuple):
        return sum(_count_leaves(child) for child in tree)
    return 1


@given(st.data())
@settings(max_examples=5, suppress_health_check=["filter_too_much"])
def test_respects_min_leaves(data):
    min_leaves = data.draw(st.integers(1, 20))
    max_leaves = data.draw(st.integers(min_leaves, 40))
    note(f"{min_leaves=}")
    note(f"{max_leaves=}")
    s = st.recursive(
        st.none(),
        lambda x: st.tuples(x, x),
        min_leaves=min_leaves,
        max_leaves=max_leaves,
    )
    assert_all_examples(s, lambda tree: min_leaves <= _count_leaves(tree) <= max_leaves)


@given(st.recursive(st.none(), lambda x: st.tuples(x, x), min_leaves=5, max_leaves=5))
@settings(suppress_health_check=["filter_too_much"])
def test_can_set_exact_leaf_count(tree):
    assert _count_leaves(tree) == 5


def test_identity_extend_warns():
    with pytest.warns(HypothesisWarning, match="extend=lambda x: x is a no-op"):
        st.recursive(st.none(), lambda x: x).validate()



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import pytest

from hypothesis import given
from hypothesis.errors import Flaky, FlakyFailure
from hypothesis.strategies import composite, integers, none

from tests.common.utils import Why, skipif_threading, xfail_on_crosshair


@pytest.mark.parametrize(
    "e", [KeyboardInterrupt, SystemExit, GeneratorExit, ValueError]
)
def test_exception_propagates_fine(e):
    @given(integers())
    def test_raise(x):
        raise e

    with pytest.raises(e):
        test_raise()


@pytest.mark.parametrize(
    "e", [KeyboardInterrupt, SystemExit, GeneratorExit, ValueError]
)
def test_exception_propagates_fine_from_strategy(e):
    @composite
    def interrupt_eventually(draw):
        raise e
        # this line will not be executed, but must be here
        # to pass draw function static reference check
        return draw(none())

    @given(interrupt_eventually())
    def test_do_nothing(x):
        pass

    with pytest.raises(e):
        test_do_nothing()


@xfail_on_crosshair(Why.other, strict=False)  # extra replay from backend switch
@pytest.mark.parametrize("e", [KeyboardInterrupt, ValueError])
def test_baseexception_no_rerun_no_flaky(e):
    runs = 0
    interrupt = 3

    @given(integers())
    def test_raise_baseexception(x):
        nonlocal runs
        runs += 1
        if runs == interrupt:
            raise e

    if issubclass(e, (KeyboardInterrupt, SystemExit, GeneratorExit)):
        # Here SystemExit and GeneratorExit are passed through
        with pytest.raises(e):
            test_raise_baseexception()

        assert runs == interrupt
    else:
        with pytest.raises(FlakyFailure):
            test_raise_baseexception()


@xfail_on_crosshair(Why.symbolic_outside_context, strict=False)  # KI and GE only
@pytest.mark.parametrize(
    "e", [KeyboardInterrupt, SystemExit, GeneratorExit, ValueError]
)
def test_baseexception_in_strategy_no_rerun_no_flaky(e):
    runs = 0
    interrupt = 3

    @composite
    def interrupt_eventually(draw):
        nonlocal runs
        runs += 1
        if runs == interrupt:
            raise e
        return draw(integers())

    @given(interrupt_eventually())
    def test_do_nothing(x):
        pass

    if issubclass(e, KeyboardInterrupt):
        with pytest.raises(e):
            test_do_nothing()

        assert runs == interrupt

    else:
        # Now SystemExit and GeneratorExit are caught like other exceptions
        with pytest.raises(Flaky):
            test_do_nothing()


TEMPLATE = """
from hypothesis import given, note, strategies as st

@st.composite
def things(draw):
    raise {exception}
    # this line will not be executed, but must be here
    # to pass draw function static reference check
    return draw(st.none())


@given(st.data(), st.integers())
def test(data, x):
    if x > 100:
        data.draw({strategy})
        raise {exception}
"""


@skipif_threading  # something in pytest here is not thread safe
@pytest.mark.parametrize("exc_name", ["SystemExit", "GeneratorExit"])
@pytest.mark.parametrize("use_composite", [True, False])
def test_explanations(testdir, exc_name, use_composite):
    code = TEMPLATE.format(
        exception=exc_name, strategy="things()" if use_composite else "st.none()"
    )
    test_file = str(testdir.makepyfile(code))
    pytest_stdout = str(testdir.runpytest_inprocess(test_file, "--tb=native").stdout)
    assert "x=101" in pytest_stdout
    assert exc_name in pytest_stdout



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import sys
from contextlib import contextmanager

import pytest

from hypothesis import given, strategies as st
from hypothesis.errors import HypothesisWarning
from hypothesis.internal.scrutineer import MONITORING_TOOL_ID


@contextmanager
def using_tool_id(tool_id, tool_name):
    try:
        sys.monitoring.use_tool_id(tool_id, tool_name)
        yield
    finally:
        sys.monitoring.free_tool_id(tool_id)


@pytest.mark.skipif(sys.version_info[:2] < (3, 12), reason="new namespace")
def test_monitoring_warns_on_registered_tool_id(warns_or_raises):

    # scrutineer can't run if something has already registered its tool id.
    with (
        using_tool_id(MONITORING_TOOL_ID, "rogue"),
        warns_or_raises(HypothesisWarning, match=r"already taken by tool rogue"),
    ):

        @given(st.integers())
        def f(n):
            raise AssertionError

        with pytest.raises(AssertionError):
            f()



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

from hypothesis import HealthCheck, Verbosity, assume, given, settings, strategies as st


@settings(max_examples=1, database=None)
@given(st.integers())
def test_single_example(n):
    pass


@settings(
    max_examples=1,
    database=None,
    suppress_health_check=[HealthCheck.filter_too_much, HealthCheck.too_slow],
    verbosity=Verbosity.debug,
)
@given(st.integers())
def test_hard_to_find_single_example(n):
    # Numbers are arbitrary, just deliberately unlikely to hit this too soon.
    assume(n % 50 == 11)



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import inspect
import math
from copy import copy

import pytest

from hypothesis import HealthCheck, assume, given, settings, strategies as st
from hypothesis.internal.compat import ExceptionGroup
from hypothesis.internal.conjecture.data import ConjectureData
from hypothesis.strategies._internal.random import (
    RANDOM_METHODS,
    ArtificialRandom,
    HypothesisRandom,
    TrueRandom,
    convert_kwargs,
    normalize_zero,
)

from tests.common.debug import assert_all_examples, find_any
from tests.common.utils import Why, xfail_on_crosshair


def test_implements_all_random_methods():
    for name in dir(HypothesisRandom):
        if not name.startswith("_") or name == "_randbelow":
            f = getattr(HypothesisRandom, name)
            if inspect.isfunction(f):
                assert f.__module__ == "hypothesis.strategies._internal.random", name


any_random = st.randoms(use_true_random=False) | st.randoms(use_true_random=True)

beta_param = st.floats(0.01, 1000)
seq_param = st.lists(st.integers(), min_size=1)


METHOD_STRATEGIES = {}


def define_method_strategy(name, **kwargs):
    METHOD_STRATEGIES[name] = kwargs


define_method_strategy("betavariate", alpha=beta_param, beta=beta_param)
define_method_strategy("binomialvariate", n=st.integers(min_value=1), p=st.floats(0, 1))
define_method_strategy("gammavariate", alpha=beta_param, beta=beta_param)
define_method_strategy("weibullvariate", alpha=beta_param, beta=beta_param)
define_method_strategy("choice", seq=seq_param)
define_method_strategy("choices", population=seq_param, k=st.integers(1, 100))
define_method_strategy("expovariate", lambd=beta_param)
define_method_strategy("_randbelow", n=st.integers(1, 2**64))
define_method_strategy("random")
define_method_strategy("getrandbits", n=st.integers(1, 128))
define_method_strategy("gauss", mu=st.floats(-1000, 1000), sigma=beta_param)
define_method_strategy("normalvariate", mu=st.floats(-1000, 1000), sigma=beta_param)
# the standard library lognormalvariate is weirdly bad at handling large floats
define_method_strategy(
    "lognormvariate", mu=st.floats(0.1, 10), sigma=st.floats(0.1, 10)
)
define_method_strategy(
    "vonmisesvariate", mu=st.floats(0, math.pi * 2), kappa=beta_param
)
# Small alpha may raise ZeroDivisionError, see https://bugs.python.org/issue41421
define_method_strategy("paretovariate", alpha=st.floats(min_value=1.0))
define_method_strategy("shuffle", x=st.lists(st.integers()))
define_method_strategy("randbytes", n=st.integers(0, 100))


INT64 = st.integers(-(2**63), 2**63 - 1)


@st.composite
def any_call_of_method(draw, method):
    if method == "sample":
        population = draw(seq_param)
        k = draw(st.integers(0, len(population)))
        kwargs = {"population": population, "k": k}
    elif method == "randint":
        a = draw(INT64)
        b = draw(INT64)
        a, b = sorted((a, b))
        kwargs = {"a": a, "b": b}
    elif method == "randrange":
        a = draw(INT64)
        b = draw(INT64)
        assume(a != b)
        a, b = sorted((a, b))
        kwargs = {"start": a, "stop": b, "step": draw(st.integers(1, 3))}
    elif method == "triangular":
        a = normalize_zero(draw(st.floats(allow_infinity=False, allow_nan=False)))
        b = normalize_zero(draw(st.floats(allow_infinity=False, allow_nan=False)))
        a, b = sorted((a, b))
        if draw(st.booleans()):
            draw(st.floats(a, b))
        kwargs = {"low": a, "high": b, "mode": None}
    elif method == "uniform":
        a = normalize_zero(draw(st.floats(allow_infinity=False, allow_nan=False)))
        b = normalize_zero(draw(st.floats(allow_infinity=False, allow_nan=False)))
        a, b = sorted((a, b))
        kwargs = {"a": a, "b": b}
    else:
        kwargs = draw(st.fixed_dictionaries(METHOD_STRATEGIES[method]))

    args, kwargs = convert_kwargs(method, kwargs)

    return (args, kwargs)


@st.composite
def any_call(draw):
    method = draw(st.sampled_from(RANDOM_METHODS))
    return (method, *draw(any_call_of_method(method)))


@pytest.mark.parametrize("method", RANDOM_METHODS)
@given(any_random, st.data())
def test_call_all_methods(method, rnd, data):
    args, kwargs = data.draw(any_call_of_method(method))
    getattr(rnd, method)(*args, **kwargs)


@given(any_random, st.integers(1, 100))
def test_rand_below(rnd, n):
    assert rnd._randbelow(n) < n


@given(any_random, beta_param, beta_param)
def test_beta_in_range(rnd, a, b):
    assert 0 <= rnd.betavariate(a, b) <= 1


def test_multiple_randoms_are_unrelated():
    @given(st.randoms(use_true_random=False), st.randoms(use_true_random=False))
    def test(r1, r2):
        assert r1.random() == r2.random()

    with pytest.raises(AssertionError):
        test()


@pytest.mark.parametrize("use_true_random", [False, True])
@given(data=st.data())
def test_randoms_can_be_synced(use_true_random, data):
    r1 = data.draw(st.randoms(use_true_random=use_true_random))
    r2 = data.draw(st.randoms(use_true_random=use_true_random))
    r2.setstate(r1.getstate())
    assert r1.random() == r2.random()


@pytest.mark.parametrize("use_true_random", [False, True])
@given(data=st.data(), method_call=any_call())
def test_seeding_to_same_value_synchronizes(use_true_random, data, method_call):
    r1 = data.draw(st.randoms(use_true_random=use_true_random))
    r2 = data.draw(st.randoms(use_true_random=use_true_random))
    method, args, kwargs = method_call
    r1.seed(0)
    r2.seed(0)
    assert getattr(r1, method)(*args, **kwargs) == getattr(r2, method)(*args, **kwargs)


@given(any_random, any_call())
def test_copying_synchronizes(r1, method_call):
    method, args, kwargs = method_call
    r2 = copy(r1)
    assert getattr(r1, method)(*args, **kwargs) == getattr(r2, method)(*args, **kwargs)


@xfail_on_crosshair(Why.symbolic_outside_context, strict=False)
@pytest.mark.parametrize("use_true_random", [True, False])
def test_seeding_to_different_values_does_not_synchronize(use_true_random):
    @given(
        st.randoms(use_true_random=use_true_random),
        st.randoms(use_true_random=use_true_random),
    )
    def test(r1, r2):
        r1.seed(0)
        r2.seed(1)
        assert r1.random() == r2.random()

    with pytest.raises(AssertionError):
        test()


@xfail_on_crosshair(Why.symbolic_outside_context, strict=False)
@pytest.mark.parametrize("use_true_random", [True, False])
def test_unrelated_calls_desynchronizes(use_true_random):
    @given(
        st.randoms(use_true_random=use_true_random),
        st.randoms(use_true_random=use_true_random),
    )
    def test(r1, r2):
        r1.seed(0)
        r2.seed(0)
        r1.randrange(1, 10)
        r2.getrandbits(128)
        assert r1.random() == r2.random()

    with pytest.raises(AssertionError):
        test()


@given(st.randoms(use_true_random=False), st.randoms(use_true_random=False))
def test_state_is_consistent(r1, r2):
    r2.setstate(r1.getstate())
    assert r1.getstate() == r2.getstate()


@given(st.randoms())
def test_does_not_use_true_random_by_default(rnd):
    assert not isinstance(rnd, TrueRandom)


@given(st.randoms(use_true_random=False))
def test_handles_singleton_uniforms_correctly(rnd):
    assert rnd.uniform(1.0, 1.0) == 1.0
    assert rnd.uniform(0.0, 0.0) == 0.0
    assert rnd.uniform(-0.0, -0.0) == 0.0
    assert rnd.uniform(0.0, -0.0) == 0.0


@given(st.randoms(use_true_random=False))
def test_handles_singleton_regions_of_triangular_correctly(rnd):
    assert rnd.triangular(1.0, 1.0) == 1.0
    assert rnd.triangular(0.0, 0.0) == 0.0
    assert rnd.triangular(-0.0, -0.0) == 0.0
    assert rnd.triangular(0.0, -0.0) == 0.0


@given(st.randoms(use_true_random=False))
def test_triangular_with_mode(rnd):
    x = rnd.triangular(0.0, 1.0, mode=0.5)
    assert 0.0 <= x <= 1.0


@pytest.mark.parametrize("use_true_random", [False, True])
def test_outputs_random_calls(use_true_random):
    @given(st.randoms(use_true_random=use_true_random, note_method_calls=True))
    def test(rnd):
        rnd.uniform(0.1, 0.5)
        raise AssertionError

    with pytest.raises(AssertionError) as err:
        test()
    assert ".uniform(0.1, 0.5)" in "\n".join(err.value.__notes__)


@pytest.mark.skipif(
    "choices" not in RANDOM_METHODS,
    reason="choices not supported on this Python version",
)
@pytest.mark.parametrize("use_true_random", [False, True])
def test_converts_kwargs_correctly_in_output(use_true_random):
    @given(st.randoms(use_true_random=use_true_random, note_method_calls=True))
    def test(rnd):
        rnd.choices([1, 2, 3, 4], k=2)
        raise AssertionError

    with pytest.raises(AssertionError) as err:
        test()
    assert ".choices([1, 2, 3, 4], k=2)" in "\n".join(err.value.__notes__)


@given(st.randoms(use_true_random=False))
def test_some_ranges_are_in_range(rnd):
    assert 0 <= rnd.randrange(10) < 10
    assert 11 <= rnd.randrange(11, 20) < 20
    assert rnd.randrange(1, 100, 3) in range(1, 100, 3)
    assert rnd.randrange(100, step=3) in range(0, 100, 3)


def test_invalid_range():
    @given(st.randoms(use_true_random=False))
    def test(rnd):
        rnd.randrange(1, 1)

    with pytest.raises(ValueError):
        test()


def test_invalid_sample():
    @given(st.randoms(use_true_random=False))
    def test(rnd):
        rnd.sample([1, 2], 3)

    with pytest.raises(ValueError):
        test()


def test_triangular_modes():
    @settings(report_multiple_bugs=True)
    @given(st.randoms(use_true_random=False))
    def test(rnd):
        x = rnd.triangular(0.0, 1.0, mode=0.5)
        assert x < 0.5
        assert x > 0.5

    with pytest.raises(ExceptionGroup):
        test()


@given(st.randoms(use_true_random=False), any_call_of_method("sample"))
def test_samples_have_right_length(rnd, sample):
    (seq, k), _ = sample
    assert len(rnd.sample(seq, k)) == k


@given(st.randoms(use_true_random=False), any_call_of_method("choices"))
def test_choices_have_right_length(rnd, choices):
    args, kwargs = choices

    seq = args[0]
    k = kwargs.get("k", 1)
    assert len(rnd.choices(seq, k=k)) == k


@given(any_random, st.integers(0, 100))
def test_randbytes_have_right_length(rnd, n):
    assert len(rnd.randbytes(n)) == n


@pytest.mark.skipif(
    settings.get_current_profile_name() == "crosshair",
    reason="takes hours; may get faster after https://github.com/pschanely/CrossHair/issues/332",
)
@given(any_random)
def test_can_manage_very_long_ranges_with_step(rnd):
    i = rnd.randrange(0, 2**256, 3)

    assert i % 3 == 0
    assert 0 <= i < 2**256
    assert i in range(0, 2**256, 3)


@settings(suppress_health_check=[HealthCheck.too_slow])
@given(any_random, st.data())
def test_range_with_arbitrary_step_is_in_range(rnd, data):
    endpoints = st.integers(-100, 100)
    step = data.draw(st.integers(1, 3))
    start, stop = sorted((data.draw(endpoints), data.draw(endpoints)))
    assume(start < stop)
    i = rnd.randrange(start, stop, step)
    assert i in range(start, stop, step)


@given(any_random, st.integers(min_value=1))
def test_range_with_only_stop(rnd, n):
    assert 0 <= rnd.randrange(n) < n


def test_can_find_end_of_range():
    find_any(
        st.randoms(use_true_random=False).map(lambda r: r.randrange(0, 11, 2)),
        lambda n: n == 10,
    )
    find_any(
        st.randoms(use_true_random=False).map(lambda r: r.randrange(0, 10, 2)),
        lambda n: n == 8,
    )


@given(st.randoms(use_true_random=False))
def test_can_sample_from_whole_range(rnd):
    xs = list(map(str, range(10)))
    ys = rnd.sample(xs, len(xs))
    assert sorted(ys) == sorted(xs)


@given(st.randoms(use_true_random=False))
def test_can_sample_from_large_subset(rnd):
    xs = list(map(str, range(10)))
    n = len(xs) // 3
    ys = rnd.sample(xs, n)
    assert set(ys).issubset(set(xs))
    assert len(ys) == len(set(ys)) == n


@given(st.randoms(use_true_random=False))
def test_can_draw_empty_from_empty_sequence(rnd):
    assert rnd.sample([], 0) == []


def test_random_includes_zero_excludes_one():
    strat = st.randoms(use_true_random=False).map(lambda r: r.random())
    assert_all_examples(strat, lambda x: 0 <= x < 1)
    find_any(strat, lambda x: x == 0)


def test_betavariate_includes_zero_and_one():
    # https://github.com/HypothesisWorks/hypothesis/issues/4297#issuecomment-2720144709
    strat = st.randoms(use_true_random=False).flatmap(
        lambda r: st.builds(
            r.betavariate, alpha=st.just(1.0) | beta_param, beta=beta_param
        )
    )
    assert_all_examples(strat, lambda x: 0 <= x <= 1)
    find_any(strat, lambda x: x == 0)
    find_any(strat, lambda x: x == 1)


def test_artificial_random_with_already_initialized_states_for_ids():
    # covering test for calling .getstate when data.states_for_ids is not None.
    data = ConjectureData.for_choices([])
    data.states_for_ids = {}
    r = ArtificialRandom(note_method_calls=False, data=data)
    r.getstate()



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

from sys import float_info

import pytest

from hypothesis.errors import InvalidArgument
from hypothesis.internal.floats import next_down, next_up
from hypothesis.strategies import floats
from hypothesis.strategies._internal.numbers import next_down_normal, next_up_normal

from tests.common.debug import assert_no_examples, check_can_generate_examples, find_any
from tests.common.utils import PYTHON_FTZ

pytestmark = [pytest.mark.skipif(PYTHON_FTZ, reason="broken by unsafe compiler flags")]


def kw(marks=(), **kwargs):
    id_ = ", ".join(f"{k}={v!r}" for k, v in kwargs.items())
    return pytest.param(kwargs, id=id_, marks=marks)


@pytest.mark.parametrize(
    "kwargs",
    [
        kw(min_value=1),
        kw(max_value=-1),
        kw(min_value=float_info.min),
        kw(min_value=next_down(float_info.min), exclude_min=True),
        kw(max_value=-float_info.min),
        kw(min_value=next_up(-float_info.min), exclude_max=True),
    ],
)
def test_subnormal_validation(kwargs):
    strat = floats(**kwargs, allow_subnormal=True)
    with pytest.raises(InvalidArgument):
        check_can_generate_examples(strat)


@pytest.mark.parametrize(
    "kwargs",
    [
        # min value
        kw(allow_subnormal=False, min_value=1),
        kw(allow_subnormal=False, min_value=float_info.min),
        kw(allow_subnormal=True, min_value=-1),
        kw(allow_subnormal=True, min_value=next_down(float_info.min)),
        # max value
        kw(allow_subnormal=False, max_value=-1),
        kw(allow_subnormal=False, max_value=-float_info.min),
        kw(allow_subnormal=True, max_value=1),
        kw(allow_subnormal=True, max_value=next_up(-float_info.min)),
        # min/max values
        kw(allow_subnormal=True, min_value=-1, max_value=1),
        kw(
            allow_subnormal=True,
            min_value=next_down(float_info.min),
            max_value=float_info.min,
        ),
        kw(
            allow_subnormal=True,
            min_value=-float_info.min,
            max_value=next_up(-float_info.min),
        ),
        kw(allow_subnormal=False, min_value=-1, max_value=-float_info.min),
        kw(allow_subnormal=False, min_value=float_info.min, max_value=1),
    ],
)
def test_allow_subnormal_defaults_correctly(kwargs):
    # copy to support our threading CI tests
    kwargs = kwargs.copy()
    allow_subnormal = kwargs.pop("allow_subnormal")
    strat = floats(**kwargs).filter(lambda x: x != 0)
    if allow_subnormal:
        find_any(strat, lambda x: -float_info.min < x < float_info.min)
    else:
        assert_no_examples(strat, lambda x: -float_info.min < x < float_info.min)


@pytest.mark.parametrize(
    "func, val, expected",
    [
        (next_up_normal, -float_info.min, -0.0),
        (next_up_normal, +0.0, float_info.min),
        (next_down_normal, float_info.min, +0.0),
        (next_down_normal, -0.0, -float_info.min),
    ],
)
def test_next_float_normal(func, val, expected):
    assert func(value=val, width=64, allow_subnormal=False) == expected



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import abc
import contextlib
import math
import sys
import warnings
from collections.abc import Iterable
from contextlib import AbstractContextManager, contextmanager
from functools import cached_property
from random import Random
from sys import float_info
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Literal,
    Optional,
    TypeAlias,
    TypedDict,
    TypeVar,
)

from sortedcontainers import SortedSet

from hypothesis.errors import HypothesisWarning
from hypothesis.internal.cache import LRUCache
from hypothesis.internal.compat import WINDOWS, int_from_bytes
from hypothesis.internal.conjecture.choice import (
    ChoiceConstraintsT,
    ChoiceT,
    ChoiceTypeT,
    FloatConstraints,
    choice_constraints_key,
    choice_permitted,
)
from hypothesis.internal.conjecture.floats import lex_to_float
from hypothesis.internal.conjecture.junkdrawer import bits_to_bytes
from hypothesis.internal.conjecture.utils import (
    INT_SIZES,
    INT_SIZES_SAMPLER,
    Sampler,
    many,
)
from hypothesis.internal.constants_ast import (
    Constants,
    constants_from_module,
    is_local_module_file,
)
from hypothesis.internal.floats import (
    SIGNALING_NAN,
    float_to_int,
    make_float_clamper,
    next_down,
    next_up,
)
from hypothesis.internal.intervalsets import IntervalSet
from hypothesis.internal.observability import InfoObservationType, TestCaseObservation

if TYPE_CHECKING:
    from hypothesis.internal.conjecture.data import ConjectureData
    from hypothesis.internal.constants_ast import ConstantT

T = TypeVar("T")
LifetimeT: TypeAlias = Literal["test_case", "test_function"]
COLLECTION_DEFAULT_MAX_SIZE = 10**10  # "arbitrarily large"


#: Registered Hypothesis backends. This is a dictionary where keys are the name
#: to be used in |settings.backend|. The value of a key can be either:
#:
#: * A string corresponding to an importable absolute path of a
#:   |PrimitiveProvider| subclass
#: * A |PrimitiveProvider| subclass (the class itself, not an instance of the
#:   class)
#:
#: Hypothesis will instantiate the corresponding |PrimitiveProvider| subclass
#: when the backend is requested by a test's |settings.backend| value.
#:
#: For example, the default Hypothesis backend is registered as:
#:
#: .. code-block:: python
#:
#:    from hypothesis.internal.conjecture.providers import AVAILABLE_PROVIDERS
#:
#:    AVAILABLE_PROVIDERS["hypothesis"] = "hypothesis.internal.conjecture.providers.HypothesisProvider"
#:    # or
#:    AVAILABLE_PROVIDERS["hypothesis"] = HypothesisProvider
#:
#: And can be used with:
#:
#: .. code-block:: python
#:
#:     from hypothesis import given, settings, strategies as st
#:
#:     @given(st.integers())
#:     @settings(backend="hypothesis")
#:     def f(n):
#:         pass
#:
#: Though, as ``backend="hypothesis"`` is the default setting, the above would
#: typically not have any effect.
#:
#: For third-party backend authors, we strongly encourage ensuring that
#: ``import hypothesis`` does not automatically import the expensive parts of
#: your package, by:
#:
#: - setting a string path here, instead of a provider class
#: - ensuring the registered hypothesis plugin path references a path which just
#:   sets AVAILABLE_PROVIDERS and does not import your package
AVAILABLE_PROVIDERS: dict[str, str | type["PrimitiveProvider"]] = {
    "hypothesis": "hypothesis.internal.conjecture.providers.HypothesisProvider",
    "hypothesis-urandom": "hypothesis.internal.conjecture.providers.URandomProvider",
}
# cache the choice_permitted constants for a particular set of constraints.
CacheKeyT: TypeAlias = tuple[ChoiceTypeT, tuple[Any, ...]]
CacheValueT: TypeAlias = tuple[tuple["ConstantT", ...], tuple["ConstantT", ...]]
CONSTANTS_CACHE: LRUCache[CacheKeyT, CacheValueT] = LRUCache(1024)


_constants_integers = (
    # powers of 2
    [2**n for n in range(16, 66)]
    # powers of 10
    + [10**n for n in range(5, 20)]
    # factorials
    + [math.factorial(n) for n in range(9, 21)]
    # a few primorial numbers https://en.wikipedia.org/wiki/Primorial
    + [
        510510,
        6469693230,
        304250263527210,
        32589158477190044730,
    ]
)
_constants_integers.extend(
    [n - 1 for n in _constants_integers] + [n + 1 for n in _constants_integers]
)
_constants_integers.extend([-x for x in _constants_integers])

# arbitrary cutoffs to keep our list bounded
assert all(50_000 <= abs(n) <= 2**66 for n in _constants_integers)

_constant_floats = (
    [
        0.5,
        1.1,
        1.5,
        1.9,
        1.0 / 3,
        10e6,
        10e-6,
        1.175494351e-38,
        next_up(0.0),
        float_info.min,
        float_info.max,
        3.402823466e38,
        9007199254740992.0,
        1 - 10e-6,
        2 + 10e-6,
        1.192092896e-07,
        2.2204460492503131e-016,
    ]
    + [2.0**-n for n in (24, 14, 149, 126)]  # minimum (sub)normals for float16,32
    + [float_info.min / n for n in (2, 10, 1000, 100_000)]  # subnormal in float64
)
_constant_floats.extend([-x for x in _constant_floats])
assert all(isinstance(f, float) for f in _constant_floats)

_constant_strings = {
    # strings which can be interpreted as code / logic
    "undefined",
    "null",
    "NULL",
    "nil",
    "NIL",
    "true",
    "false",
    "True",
    "False",
    "TRUE",
    "FALSE",
    "None",
    "none",
    "if",
    "then",
    "else",
    "__dict__",
    "__proto__",  # javascript
    # strings which can be interpreted as a number
    "0",
    "1e100",
    "0..0",
    "0/0",
    "1/0",
    "+0.0",
    "Infinity",
    "-Infinity",
    "Inf",
    "INF",
    "NaN",
    "9" * 30,
    # common ascii characters
    ",./;'[]\\-=<>?:\"{}|_+!@#$%^&*()`~",
    # common unicode characters
    "Ω≈ç√∫˜µ≤≥÷åß∂ƒ©˙∆˚¬…æœ∑´®†¥¨ˆøπ“‘¡™£¢∞§¶•ªº–≠¸˛Ç◊ı˜Â¯˘¿ÅÍÎÏ˝ÓÔÒÚÆ☃Œ„´‰ˇÁ¨ˆØ∏”’`⁄€‹›ﬁﬂ‡°·‚—±",
    # characters which increase in length when lowercased
    "Ⱥ",
    "Ⱦ",
    # ligatures
    "æœÆŒﬀʤʨß",
    # emoticons
    "(╯°□°）╯︵ ┻━┻)",
    # emojis
    "😍",
    "🇺🇸",
    # emoji modifiers
    "🏻",  # U+1F3FB Light Skin Tone,
    "👍🏻",  # 👍 followed by U+1F3FB
    # RTL text
    "الكل في المجمو عة",
    # Ogham text, which contains the only character in the Space Separators
    # unicode category (Zs) that isn't visually blank:  .  # noqa: RUF003
    "᚛ᚄᚓᚐᚋᚒᚄ ᚑᚄᚂᚑᚏᚅ᚜",
    # thai consonant + spacing vowel combinations, which have unusual visual combining behavior
    "กา",
    "ก ำกำ",
    # readable variations on text (bolt/italic/script)
    "𝐓𝐡𝐞 𝐪𝐮𝐢𝐜𝐤 𝐛𝐫𝐨𝐰𝐧 𝐟𝐨𝐱 𝐣𝐮𝐦𝐩𝐬 𝐨𝐯𝐞𝐫 𝐭𝐡𝐞 𝐥𝐚𝐳𝐲 𝐝𝐨𝐠",
    "𝕿𝖍𝖊 𝖖𝖚𝖎𝖈𝖐 𝖇𝖗𝖔𝖜𝖓 𝖋𝖔𝖝 𝖏𝖚𝖒𝖕𝖘 𝖔𝖛𝖊𝖗 𝖙𝖍𝖊 𝖑𝖆𝖟𝖞 𝖉𝖔𝖌",
    "𝑻𝒉𝒆 𝒒𝒖𝒊𝒄𝒌 𝒃𝒓𝒐𝒘𝒏 𝒇𝒐𝒙 𝒋𝒖𝒎𝒑𝒔 𝒐𝒗𝒆𝒓 𝒕𝒉𝒆 𝒍𝒂𝒛𝒚 𝒅𝒐𝒈",
    "𝓣𝓱𝓮 𝓺𝓾𝓲𝓬𝓴 𝓫𝓻𝓸𝔀𝓷 𝓯𝓸𝔁 𝓳𝓾𝓶𝓹𝓼 𝓸𝓿𝓮𝓻 𝓽𝓱𝓮 𝓵𝓪𝔃𝔂 𝓭𝓸𝓰",
    "𝕋𝕙𝕖 𝕢𝕦𝕚𝕔𝕜 𝕓𝕣𝕠𝕨𝕟 𝕗𝕠𝕩 𝕛𝕦𝕞𝕡𝕤 𝕠𝕧𝕖𝕣 𝕥𝕙𝕖 𝕝𝕒𝕫𝕪 𝕕𝕠𝕘",
    # upsidown text
    "ʇǝɯɐ ʇᴉs ɹolop ɯnsdᴉ ɯǝɹo˥",
    # reserved strings in windows
    "NUL",
    "COM1",
    "LPT1",
    # scunthorpe problem
    "Scunthorpe",
    # zalgo text
    "Ṱ̺̺̕o͞ ̷i̲̬͇̪͙n̝̗͕v̟̜̘̦͟o̶̙̰̠kè͚̮̺̪̹̱̤ ̖t̝͕̳̣̻̪͞h̼͓̲̦̳̘̲e͇̣̰̦̬͎ ̢̼̻̱̘h͚͎͙̜̣̲ͅi̦̲̣̰̤v̻͍e̺̭̳̪̰-m̢iͅn̖̺̞̲̯̰d̵̼̟͙̩̼̘̳ ̞̥̱̳̭r̛̗̘e͙p͠r̼̞̻̭̗e̺̠̣͟s̘͇̳͍̝͉e͉̥̯̞̲͚̬͜ǹ̬͎͎̟̖͇̤t͍̬̤͓̼̭͘ͅi̪̱n͠g̴͉ ͏͉ͅc̬̟h͡a̫̻̯͘o̫̟̖͍̙̝͉s̗̦̲.̨̹͈̣",
    #
    # examples from https://faultlore.com/blah/text-hates-you/
    "मनीष منش",
    "पन्ह पन्ह त्र र्च कृकृ ड्ड न्हृे إلا بسم الله",
    "lorem لا بسم الله ipsum 你好1234你好",
    # unicode charaacters causing unconditional line breaks, as defined by UAX #14:
    # https://www.unicode.org/reports/tr14/.
    #
    # We've seen multiple bugs caused by assuming `str.splitlines` is equivalent to
    # splitting over "\n", while it actually splits over all line breaks!
    #
    # We intersperse the line breaks with normal characters to increase the likelihood
    # of triggering such a bug.
    (
        "a"
        "\u000a"  # line feed (class: LF)
        "b"
        "\u000d"  # carriage return (class: CR)
        "c"
        "\u0085"  # next line (class: NL)
        "d"
        "\u000b"  # line tabulation (class: BK)
        "e"
        "\u000c"  # form feed (class: BK)
        "f"
        "\u2028"  # line separator (class: BK)
        "g"
        "\u2029"  # paragraph separator (class: BK)
        "h"
        "\u000d\u000a"  # CR+LF
        "i"
    ),
}


# we don't actually care what order the constants are sorted in, just that the
# ordering is deterministic.
GLOBAL_CONSTANTS = Constants(
    integers=SortedSet(_constants_integers),
    floats=SortedSet(_constant_floats, key=float_to_int),
    bytes=SortedSet(),
    strings=SortedSet(_constant_strings),
)

_local_constants = Constants(
    integers=SortedSet(),
    floats=SortedSet(key=float_to_int),
    bytes=SortedSet(),
    strings=SortedSet(),
)
# modules that we've already seen and processed for local constants. These are
# are all modules, not necessarily local ones. This lets us quickly see which
# modules are new without an expensive path.resolve() or is_local_module_file
# cache lookup.
# We track by module object when hashable, falling back to the module name
# (str key in sys.modules) for unhashable entries like SimpleNamespace.
_seen_modules: set = set()
_sys_modules_len: int | None = None


def _get_local_constants() -> Constants:
    global _sys_modules_len, _local_constants

    if sys.platform == "emscripten":  # pragma: no cover
        # pyodide builds bundle the stdlib in a nonstandard location, like
        # `/lib/python312.zip/heapq.py`. To avoid identifying the entirety of
        # the stdlib as local code and slowing down on emscripten, instead return
        # that nothing is local.
        #
        # pyodide may provide some way to distinguish stdlib/third-party/local
        # code. I haven't looked into it. If they do, we should correctly implement
        # ModuleLocation for pyodide instead of this.
        return _local_constants

    count_constants = len(_local_constants)
    # We call this function once per HypothesisProvider instance, i.e. once per
    # input, so it needs to be performant. The logic here is more complicated
    # than necessary because of this.
    #
    # First, we check whether there are any new modules with a very cheap length
    # check. This check can be fooled if a module is added while another module is
    # removed, but the more correct check against tuple(sys.modules.keys()) is
    # substantially more expensive. Such a new module would eventually be discovered
    # if / when the length changes again in the future.
    #
    # If the length has changed, we find just modules we haven't seen before. Of
    # those, we find the ones which correspond to local modules, and extract their
    # constants.

    # careful: store sys.modules length when we first check to avoid race conditions
    # with other threads loading a module before we set _sys_modules_len.
    if (sys_modules_len := len(sys.modules)) != _sys_modules_len:
        new_modules = []
        for name, module in list(sys.modules.items()):
            try:
                seen = module in _seen_modules
            except TypeError:
                # unhashable module (e.g. SimpleNamespace); fall back to name
                seen = name in _seen_modules
            if not seen:
                new_modules.append((name, module))
        # Repeated SortedSet unions are expensive. Do the initial unions on a
        # set(), then do a one-time union with _local_constants after.
        new_constants = Constants()
        for name, module in new_modules:
            if (
                module_file := getattr(module, "__file__", None)
            ) is not None and is_local_module_file(module_file):
                new_constants |= constants_from_module(module)
            try:
                _seen_modules.add(module)
            except TypeError:
                _seen_modules.add(name)
        _local_constants |= new_constants
        _sys_modules_len = sys_modules_len

    # if we add any new constant, invalidate the constant cache for permitted values.
    # A more efficient approach would be invalidating just the keys with this
    # choice_type.
    if len(_local_constants) > count_constants:
        CONSTANTS_CACHE.cache.clear()

    return _local_constants


@contextmanager
def with_register_backend(name, provider_cls):
    try:
        AVAILABLE_PROVIDERS[name] = provider_cls
        yield
    finally:
        del AVAILABLE_PROVIDERS[name]


class _BackendInfoMsg(TypedDict):
    type: InfoObservationType
    title: str
    content: str | dict[str, Any]


# TODO_DOCS: link to choice sequence explanation page


class PrimitiveProvider(abc.ABC):
    """
    |PrimitiveProvider| is the implementation interface of a
    :ref:`Hypothesis backend <alternative-backends>`.

    A |PrimitiveProvider| is required to implement the following five
    ``draw_*`` methods:

    * |PrimitiveProvider.draw_integer|
    * |PrimitiveProvider.draw_boolean|
    * |PrimitiveProvider.draw_float|
    * |PrimitiveProvider.draw_string|
    * |PrimitiveProvider.draw_bytes|

    Each strategy in Hypothesis generates values by drawing a series of choices
    from these five methods. By overriding them, a |PrimitiveProvider| can control
    the distribution of inputs generated by Hypothesis.

    For example, :pypi:`hypothesis-crosshair` implements a |PrimitiveProvider|
    which uses an SMT solver to generate inputs that uncover new branches.

    Once you implement a |PrimitiveProvider|, you can make it available for use
    through |AVAILABLE_PROVIDERS|.
    """

    #: The lifetime of a |PrimitiveProvider| instance. Either ``test_function``
    #: or ``test_case``.
    #:
    #: If ``test_function`` (the default), a single provider instance will be
    #: instantiated and used for the entirety of each test function (i.e., roughly
    #: one provider per |@given| annotation). This can be useful for tracking state
    #: over the entirety of a test function.
    #:
    #: If ``test_case``, a new provider instance will be instantiated and used for
    #: each input Hypothesis generates.
    #:
    #: The ``conjecturedata`` argument to ``PrimitiveProvider.__init__`` will
    #: be ``None`` for a lifetime of ``test_function``, and an instance of
    #: ``ConjectureData`` for a lifetime of ``test_case``.
    #:
    #: Third-party providers likely want to set a lifetime of ``test_function``.
    lifetime: ClassVar[LifetimeT] = "test_function"

    #: Solver-based backends such as ``hypothesis-crosshair`` use symbolic values
    #: which record operations performed on them in order to discover new paths.
    #: If ``avoid_realization`` is set to ``True``, hypothesis will avoid interacting
    #: with symbolic choices returned by the provider in any way that would force
    #: the solver to narrow the range of possible values for that symbolic.
    #:
    #: Setting this to ``True`` disables some hypothesis features and optimizations.
    #: Only set this to ``True`` if it is necessary for your backend.
    avoid_realization: ClassVar[bool] = False

    #: If ``True``, |PrimitiveProvider.on_observation| will be added as a
    #: callback via |add_observability_callback|, enabling observability during
    # the lifetime of this provider. If ``False``, |PrimitiveProvider.on_observation|
    #: will never be called by Hypothesis.
    #:
    #: The opt-in behavior of observability is because enabling observability
    #: might increase runtime or memory usage.
    add_observability_callback: ClassVar[bool] = False

    def __init__(self, conjecturedata: Optional["ConjectureData"], /) -> None:
        self._cd = conjecturedata

    @abc.abstractmethod
    def draw_boolean(
        self,
        p: float = 0.5,
    ) -> bool:
        """
        Draw a boolean choice.

        Parameters
        ----------
        p: float
            The probability of returning ``True``. Between 0 and 1 inclusive.

            Except for ``0`` and ``1``, the value of ``p`` is a hint provided by
            Hypothesis, and may be ignored by the backend.

            If ``0``, the provider must return ``False``. If ``1``, the provider
            must return ``True``.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def draw_integer(
        self,
        min_value: int | None = None,
        max_value: int | None = None,
        *,
        weights: dict[int, float] | None = None,
        shrink_towards: int = 0,
    ) -> int:
        """
        Draw an integer choice.

        Parameters
        ----------
        min_value : int | None
            (Inclusive) lower bound on the integer value. If ``None``, there is
            no lower bound.
        max_value : int | None
            (Inclusive) upper bound on the integer value. If ``None``, there is
            no upper bound.
        weights: dict[int, float] | None
            Maps keys in the range [``min_value``, ``max_value``] to the probability
            of returning that key.
        shrink_towards: int
            The integer to shrink towards. This is not used during generation and
            can be ignored by backends.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def draw_float(
        self,
        *,
        min_value: float = -math.inf,
        max_value: float = math.inf,
        allow_nan: bool = True,
        smallest_nonzero_magnitude: float,
    ) -> float:
        """
        Draw a float choice.

        Parameters
        ----------
        min_value : float
            (Inclusive) lower bound on the float value.
        max_value : float
            (Inclusive) upper bound on the float value.
        allow_nan : bool
            If ``False``, it is invalid to return ``math.nan``.
        smallest_nonzero_magnitude : float
            The smallest allowed nonzero magnitude. ``draw_float`` should not
            return a float ``f`` if ``abs(f) < smallest_nonzero_magnitude``.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def draw_string(
        self,
        intervals: IntervalSet,
        *,
        min_size: int = 0,
        max_size: int = COLLECTION_DEFAULT_MAX_SIZE,
    ) -> str:
        """
        Draw a string choice.

        Parameters
        ----------
        intervals : IntervalSet
            The set of codepoints to sample from.
        min_size : int
            (Inclusive) lower bound on the string length.
        max_size : int
            (Inclusive) upper bound on the string length.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def draw_bytes(
        self,
        min_size: int = 0,
        max_size: int = COLLECTION_DEFAULT_MAX_SIZE,
    ) -> bytes:
        """
        Draw a bytes choice.

        Parameters
        ----------
        min_size : int
            (Inclusive) lower bound on the bytes length.
        max_size : int
            (Inclusive) upper bound on the bytes length.
        """
        raise NotImplementedError

    def per_test_case_context_manager(self) -> AbstractContextManager:
        """
        Returns a context manager which will be entered each time Hypothesis
        starts generating and executing one test case, and exited when that test
        case finishes generating and executing, including if any exception is
        thrown.

        In the lifecycle of a Hypothesis test, this is called before
        generating strategy values for each test case. This is just before any
        :ref:`custom executor <custom-function-execution>` is called.

        Even if not returning a custom context manager, |PrimitiveProvider|
        subclasses are welcome to override this method to know when Hypothesis
        starts and ends the execution of a single test case.
        """
        return contextlib.nullcontext()

    def realize(self, value: T, *, for_failure: bool = False) -> T:
        """
        Called whenever hypothesis requires a concrete (non-symbolic) value from
        a potentially symbolic value. Hypothesis will not check that ``value`` is
        symbolic before calling ``realize``, so you should handle the case where
        ``value`` is non-symbolic.

        The returned value should be non-symbolic.  If you cannot provide a value,
        raise |BackendCannotProceed| with a value of ``"discard_test_case"``.

        If ``for_failure`` is ``True``, the value is associated with a failing example.
        In this case, the backend should spend substantially more effort when
        attempting to realize the value, since it is important to avoid discarding
        failing examples. Backends may still raise |BackendCannotProceed| when
        ``for_failure`` is ``True``, if realization is truly impossible or if
        realization takes significantly longer than expected (say, 5 minutes).
        """
        return value

    def replay_choices(self, choices: tuple[ChoiceT, ...]) -> None:
        """
        Called when Hypothesis has discovered a choice sequence which the provider
        may wish to enqueue to replay under its own instrumentation when we next
        ask to generate a test case, rather than generating one from scratch.

        This is used to e.g. warm-start :pypi:`hypothesis-crosshair` with a corpus
        of high-code-coverage inputs discovered by
        `HypoFuzz <https://hypofuzz.com/>`_.
        """
        return None

    def observe_test_case(self) -> dict[str, Any]:
        """Called at the end of the test case when :ref:`observability
        <observability>` is enabled.

        The return value should be a non-symbolic json-encodable dictionary,
        and will be included in observations as ``observation["metadata"]["backend"]``.
        """
        return {}

    def observe_information_messages(
        self, *, lifetime: LifetimeT
    ) -> Iterable[_BackendInfoMsg]:
        """Called at the end of each test case and again at end of the test function.

        Return an iterable of ``{type: info/alert/error, title: str, content: str | dict}``
        dictionaries to be delivered as individual information messages. Hypothesis
        adds the ``run_start`` timestamp and ``property`` name for you.
        """
        assert lifetime in ("test_case", "test_function")
        yield from []

    def on_observation(self, observation: TestCaseObservation) -> None:  # noqa: B027
        """
        Called at the end of each test case which uses this provider, with the same
        ``observation["type"] == "test_case"`` observation that is passed to
        other callbacks added via |add_observability_callback|. This method is not
        called with ``observation["type"] in {"info", "alert", "error"}``
        observations.

        .. important::

            For |PrimitiveProvider.on_observation| to be called by Hypothesis,
            |PrimitiveProvider.add_observability_callback| must be set to ``True``.

            |PrimitiveProvider.on_observation| is explicitly opt-in, as enabling
            observability might increase runtime or memory usage.

        Calls to this method are guaranteed to alternate with calls to
        |PrimitiveProvider.per_test_case_context_manager|. For example:

        .. code-block:: python

            # test function starts
            per_test_case_context_manager()
            on_observation()
            per_test_case_context_manager()
            on_observation()
            ...
            # test function ends

        Note that |PrimitiveProvider.on_observation| will not be called for test
        cases which did not use this provider during generation, for example
        during |Phase.reuse| or |Phase.shrink|, or because Hypothesis switched
        to the standard Hypothesis backend after this backend raised too many
        |BackendCannotProceed| exceptions.
        """

    def span_start(self, label: int, /) -> None:  # noqa: B027  # non-abstract noop
        """Marks the beginning of a semantically meaningful span of choices.

        Spans are a depth-first tree structure. A span is opened by a call to
        |PrimitiveProvider.span_start|, and a call to |PrimitiveProvider.span_end|
        closes the most recently opened span. So the following sequence of calls:

        .. code-block:: python

            span_start(label=1)
            n1 = draw_integer()
            span_start(label=2)
            b1 = draw_boolean()
            n2 = draw_integer()
            span_end()
            f1 = draw_float()
            span_end()

        produces the following two spans of choices:

        .. code-block::

            1: [n1, b1, n2, f1]
            2: [b1, n2]

        Hypothesis uses spans to denote "semantically meaningful" sequences of
        choices. For instance, Hypothesis opens a span for the sequence of choices
        made while drawing from each strategy. Not every span corresponds to a
        strategy; the generation of e.g. each element in |st.lists| is also marked
        with a span, among others.

        ``label`` is an opaque integer, which has no defined semantics.
        The only guarantee made by Hypothesis is that all spans with the same
        "meaning" will share the same ``label``. So all spans from the same
        strategy will share the same label, as will e.g. the spans for |st.lists|
        elements.

        Providers can track calls to |PrimitiveProvider.span_start| and
        |PrimitiveProvider.span_end| to learn something about the semantics of
        the test's choice sequence. For instance, a provider could track the depth
        of the span tree, or the number of unique labels, which says something about
        the complexity of the choices being generated. Or a provider could track
        the span tree across test cases in order to determine what strategies are
        being used in what contexts.

        It is possible for Hypothesis to start and immediately stop a span,
        without calling a ``draw_*`` method in between. These spans contain zero
        choices.

        Hypothesis will always balance the number of calls to
        |PrimitiveProvider.span_start| and |PrimitiveProvider.span_end|. A call
        to |PrimitiveProvider.span_start| will always be followed by a call to
        |PrimitiveProvider.span_end| before the end of the test case.

        |PrimitiveProvider.span_start| is called from ``ConjectureData.start_span()``
        internally.
        """

    def span_end(self, discard: bool, /) -> None:  # noqa: B027
        """Marks the end of a semantically meaningful span of choices.

        ``discard`` is ``True`` when the draw was filtered out or otherwise marked
        as unlikely to contribute to the input data as seen by the user's test.
        Note however that side effects can make this determination unsound.

        |PrimitiveProvider.span_end| is called from ``ConjectureData.stop_span()``
        internally.
        """


class HypothesisProvider(PrimitiveProvider):
    lifetime = "test_case"

    def __init__(self, conjecturedata: Optional["ConjectureData"], /):
        super().__init__(conjecturedata)
        self._random = None if self._cd is None else self._cd._random

    @cached_property
    def _local_constants(self):
        # defer computation of local constants until/if we need it
        return _get_local_constants()

    def _maybe_draw_constant(
        self,
        choice_type: ChoiceTypeT,
        constraints: ChoiceConstraintsT,
        *,
        p: float = 0.05,
    ) -> Optional["ConstantT"]:
        assert self._random is not None
        assert choice_type != "boolean"
        # check whether we even want a constant before spending time computing
        # and caching the allowed constants.
        if self._random.random() > p:
            return None

        # note: this property access results in computation being done
        assert self._local_constants is not None

        key = (choice_type, choice_constraints_key(choice_type, constraints))
        if key not in CONSTANTS_CACHE:
            CONSTANTS_CACHE[key] = (
                tuple(
                    choice
                    for choice in GLOBAL_CONSTANTS.set_for_type(choice_type)
                    if choice_permitted(choice, constraints)
                ),
                tuple(
                    choice
                    for choice in self._local_constants.set_for_type(choice_type)
                    if choice_permitted(choice, constraints)
                ),
            )

        # split constants into two pools, so we still have a good chance to draw
        # global constants even if there are many local constants.
        global_constants, local_constants = CONSTANTS_CACHE[key]
        constants_lists = ([global_constants] if global_constants else []) + (
            [local_constants] if local_constants else []
        )
        if not constants_lists:
            return None

        # At this point, we've decided to use a constant. Now we select which pool
        # to draw that constant from.
        #
        # Note that this approach has a different probability distribution than
        # attempting a random.random for both global_constants and local_constants.
        constants = self._random.choice(constants_lists)
        return self._random.choice(constants)

    def draw_boolean(
        self,
        p: float = 0.5,
    ) -> bool:
        assert self._random is not None

        if p <= 0:
            return False
        if p >= 1:
            return True

        return self._random.random() < p

    def draw_integer(
        self,
        min_value: int | None = None,
        max_value: int | None = None,
        *,
        weights: dict[int, float] | None = None,
        shrink_towards: int = 0,
    ) -> int:
        assert self._cd is not None
        if (
            constant := self._maybe_draw_constant(
                "integer",
                {
                    "min_value": min_value,
                    "max_value": max_value,
                    "weights": weights,
                    "shrink_towards": shrink_towards,
                },
            )
        ) is not None:
            assert isinstance(constant, int)
            return constant

        center = 0
        if min_value is not None:
            center = max(min_value, center)
        if max_value is not None:
            center = min(max_value, center)

        if weights is not None:
            assert min_value is not None
            assert max_value is not None

            # format of weights is a mapping of ints to p, where sum(p) < 1.
            # The remaining probability mass is uniformly distributed over
            # *all* ints (not just the unmapped ones; this is somewhat undesirable,
            # but simplifies things).
            #
            # We assert that sum(p) is strictly less than 1 because it simplifies
            # handling forced values when we can force into the unmapped probability
            # mass. We should eventually remove this restriction.
            sampler = Sampler(
                [1 - sum(weights.values()), *weights.values()], observe=False
            )
            # if we're forcing, it's easiest to force into the unmapped probability
            # mass and then force the drawn value after.
            idx = sampler.sample(self._cd)

            if idx == 0:
                return self._draw_bounded_integer(min_value, max_value)
            # implicit reliance on dicts being sorted for determinism
            return list(weights)[idx - 1]

        if min_value is None and max_value is None:
            return self._draw_unbounded_integer()

        if min_value is None:
            assert max_value is not None
            probe = max_value + 1
            while max_value < probe:
                probe = center + self._draw_unbounded_integer()
            return probe

        if max_value is None:
            assert min_value is not None
            probe = min_value - 1
            while probe < min_value:
                probe = center + self._draw_unbounded_integer()
            return probe

        return self._draw_bounded_integer(min_value, max_value)

    def draw_float(
        self,
        *,
        min_value: float = -math.inf,
        max_value: float = math.inf,
        allow_nan: bool = True,
        smallest_nonzero_magnitude: float,
    ) -> float:
        assert self._random is not None

        constraints: FloatConstraints = {
            "min_value": min_value,
            "max_value": max_value,
            "allow_nan": allow_nan,
            "smallest_nonzero_magnitude": smallest_nonzero_magnitude,
        }
        if (
            constant := self._maybe_draw_constant("float", constraints, p=0.15)
        ) is not None:
            assert isinstance(constant, float)
            return constant

        # on top of the probability to draw a constant float, we independently
        # upweight 0.0/-0.0, math.inf, -math.inf, nans, and boundary values.
        weird_floats = [
            f
            for f in [
                0.0,
                -0.0,
                math.inf,
                -math.inf,
                math.nan,
                -math.nan,
                SIGNALING_NAN,
                -SIGNALING_NAN,
                min_value,
                next_up(min_value),
                min_value + 1,
                max_value - 1,
                next_down(max_value),
                max_value,
            ]
            if choice_permitted(f, constraints)
        ]

        if weird_floats and self._random.random() < 0.05:
            return self._random.choice(weird_floats)

        clamper = make_float_clamper(
            min_value,
            max_value,
            smallest_nonzero_magnitude=smallest_nonzero_magnitude,
            allow_nan=allow_nan,
        )

        result = self._draw_float()
        if allow_nan and math.isnan(result):
            clamped = result  # pragma: no cover
        else:
            clamped = clamper(result)
        if float_to_int(clamped) != float_to_int(result) and not (
            math.isnan(result) and allow_nan
        ):
            result = clamped
        return result

    def draw_string(
        self,
        intervals: IntervalSet,
        *,
        min_size: int = 0,
        max_size: int = COLLECTION_DEFAULT_MAX_SIZE,
    ) -> str:
        assert self._cd is not None
        assert self._random is not None

        if len(intervals) == 0:
            return ""

        if (
            constant := self._maybe_draw_constant(
                "string",
                {"intervals": intervals, "min_size": min_size, "max_size": max_size},
            )
        ) is not None:
            assert isinstance(constant, str)
            return constant

        average_size = min(
            max(min_size * 2, min_size + 5),
            0.5 * (min_size + max_size),
        )

        chars = []
        elements = many(
            self._cd,
            min_size=min_size,
            max_size=max_size,
            average_size=average_size,
            observe=False,
        )
        while elements.more():
            if len(intervals) > 256:
                if self.draw_boolean(0.2):
                    i = self._random.randint(256, len(intervals) - 1)
                else:
                    i = self._random.randint(0, 255)
            else:
                i = self._random.randint(0, len(intervals) - 1)

            chars.append(intervals.char_in_shrink_order(i))

        return "".join(chars)

    def draw_bytes(
        self,
        min_size: int = 0,
        max_size: int = COLLECTION_DEFAULT_MAX_SIZE,
    ) -> bytes:
        assert self._cd is not None
        assert self._random is not None

        if (
            constant := self._maybe_draw_constant(
                "bytes", {"min_size": min_size, "max_size": max_size}
            )
        ) is not None:
            assert isinstance(constant, bytes)
            return constant

        buf = bytearray()
        average_size = min(
            max(min_size * 2, min_size + 5),
            0.5 * (min_size + max_size),
        )
        elements = many(
            self._cd,
            min_size=min_size,
            max_size=max_size,
            average_size=average_size,
            observe=False,
        )
        while elements.more():
            buf += self._random.randbytes(1)

        return bytes(buf)

    def _draw_float(self) -> float:
        assert self._random is not None

        f = lex_to_float(self._random.getrandbits(64))
        sign = 1 if self._random.getrandbits(1) else -1
        return sign * f

    def _draw_unbounded_integer(self) -> int:
        assert self._cd is not None
        assert self._random is not None

        size = INT_SIZES[INT_SIZES_SAMPLER.sample(self._cd)]

        r = self._random.getrandbits(size)
        sign = r & 1
        r >>= 1
        if sign:
            r = -r
        return r

    def _draw_bounded_integer(
        self,
        lower: int,
        upper: int,
        *,
        vary_size: bool = True,
    ) -> int:
        assert lower <= upper
        assert self._cd is not None
        assert self._random is not None

        if lower == upper:
            return lower

        bits = (upper - lower).bit_length()
        if bits > 24 and vary_size and self._random.random() < 7 / 8:
            # For large ranges, we combine the uniform random distribution
            # with a weighting scheme with moderate chance.  Cutoff at 2 ** 24 so that our
            # choice of unicode characters is uniform but the 32bit distribution is not.
            idx = INT_SIZES_SAMPLER.sample(self._cd)
            cap_bits = min(bits, INT_SIZES[idx])
            upper = min(upper, lower + 2**cap_bits - 1)
            return self._random.randint(lower, upper)

        return self._random.randint(lower, upper)


# Masks for masking off the first byte of an n-bit buffer.
# The appropriate mask is stored at position n % 8.
BYTE_MASKS = [(1 << n) - 1 for n in range(8)]
BYTE_MASKS[0] = 255


class BytestringProvider(PrimitiveProvider):
    lifetime = "test_case"

    def __init__(
        self, conjecturedata: Optional["ConjectureData"], /, *, bytestring: bytes
    ):
        super().__init__(conjecturedata)
        self.bytestring = bytestring
        self.index = 0
        self.drawn = bytearray()

    def _draw_bits(self, n):
        if n == 0:  # pragma: no cover
            return 0
        n_bytes = bits_to_bytes(n)
        if self.index + n_bytes > len(self.bytestring):
            self._cd.mark_overrun()
        buf = bytearray(self.bytestring[self.index : self.index + n_bytes])
        self.index += n_bytes

        buf[0] &= BYTE_MASKS[n % 8]
        buf = bytes(buf)
        self.drawn += buf
        return int_from_bytes(buf)

    def draw_boolean(
        self,
        p: float = 0.5,
    ) -> bool:
        if p <= 0:
            return False
        if p >= 1:
            return True

        # always use one byte for booleans to maintain constant draw size.
        # If a probability requires more than 8 bits to represent precisely,
        # the result will be slightly biased, but not badly.
        bits = 8
        size = 2**bits
        # always leave at least one value that can be true, even for very small
        # p.
        falsey = max(1, math.floor(size * (1 - p)))
        n = self._draw_bits(bits)
        return n >= falsey

    def draw_integer(
        self,
        min_value: int | None = None,
        max_value: int | None = None,
        *,
        weights: dict[int, float] | None = None,
        shrink_towards: int = 0,
    ) -> int:
        assert self._cd is not None

        # we explicitly ignore integer weights for now, as they are likely net
        # negative on fuzzer performance.

        if min_value is None and max_value is None:
            min_value = -(2**127)
            max_value = 2**127 - 1
        elif min_value is None:
            assert max_value is not None
            min_value = max_value - 2**64
        elif max_value is None:
            assert min_value is not None
            max_value = min_value + 2**64

        if min_value == max_value:
            return min_value

        bits = (max_value - min_value).bit_length()
        value = self._draw_bits(bits)
        while not (min_value <= value <= max_value):
            value = self._draw_bits(bits)
        return value

    def draw_float(
        self,
        *,
        min_value: float = -math.inf,
        max_value: float = math.inf,
        allow_nan: bool = True,
        smallest_nonzero_magnitude: float,
    ) -> float:
        n = self._draw_bits(64)
        sign = -1 if n >> 64 else 1
        f = sign * lex_to_float(n & ((1 << 64) - 1))
        clamper = make_float_clamper(
            min_value,
            max_value,
            smallest_nonzero_magnitude=smallest_nonzero_magnitude,
            allow_nan=allow_nan,
        )
        return clamper(f)

    def _draw_collection(self, min_size, max_size, *, alphabet_size):
        average_size = min(
            max(min_size * 2, min_size + 5),
            0.5 * (min_size + max_size),
        )
        elements = many(
            self._cd,
            min_size=min_size,
            max_size=max_size,
            average_size=average_size,
            observe=False,
        )
        values = []
        while elements.more():
            values.append(self.draw_integer(0, alphabet_size - 1))
        return values

    def draw_string(
        self,
        intervals: IntervalSet,
        *,
        min_size: int = 0,
        max_size: int = COLLECTION_DEFAULT_MAX_SIZE,
    ) -> str:
        values = self._draw_collection(min_size, max_size, alphabet_size=len(intervals))
        return "".join(chr(intervals[v]) for v in values)

    def draw_bytes(
        self,
        min_size: int = 0,
        max_size: int = COLLECTION_DEFAULT_MAX_SIZE,
    ) -> bytes:
        values = self._draw_collection(min_size, max_size, alphabet_size=2**8)
        return bytes(values)


class URandom(Random):
    # we reimplement a Random instance instead of using SystemRandom, because
    # os.urandom is not guaranteed to read from /dev/urandom.

    @staticmethod
    def _urandom(size: int) -> bytes:
        with open("/dev/urandom", "rb") as f:
            return f.read(size)

    def getrandbits(self, k: int) -> int:
        assert k >= 0
        size = bits_to_bytes(k)
        n = int_from_bytes(self._urandom(size))
        # trim excess bits
        return n >> (size * 8 - k)

    def random(self) -> float:
        # adapted from random.SystemRandom.random
        return (int_from_bytes(self._urandom(7)) >> 3) * (2**-53)


class URandomProvider(HypothesisProvider):
    # A provider which reads directly from /dev/urandom as its source of randomness.
    # This provider exists to provide better Hypothesis integration with Antithesis
    # (https://antithesis.com/), which interprets calls to /dev/urandom as the
    # randomness to mutate. This effectively gives Antithesis control over
    # the choices made by the URandomProvider.
    #
    # If you are not using Antithesis, you probably don't want to use this
    # provider.

    def __init__(self, conjecturedata: Optional["ConjectureData"], /):
        super().__init__(conjecturedata)
        if WINDOWS:  # pragma: no cover
            warnings.warn(
                "/dev/urandom is not available on windows. Falling back to "
                'standard PRNG generation (equivalent to backend="hypothesis").',
                HypothesisWarning,
                stacklevel=1,
            )
            # don't overwrite the HypothesisProvider self._random attribute in
            # this case
        else:
            self._random = URandom()



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import contextlib
import enum
import math
import sys
import time
import warnings
from io import StringIO
from threading import Barrier, Lock, RLock, Thread

import pytest
from pytest import mark

from hypothesis import Phase, settings
from hypothesis.errors import HypothesisDeprecationWarning
from hypothesis.internal import observability
from hypothesis.internal.floats import next_down
from hypothesis.internal.observability import (
    Observation,
    add_observability_callback,
    remove_observability_callback,
)
from hypothesis.internal.reflection import get_pretty_function_description, proxies
from hypothesis.reporting import default, with_reporter
from hypothesis.strategies._internal.core import from_type, register_type_strategy
from hypothesis.strategies._internal.types import _global_type_lookup

# we need real time here, not monkeypatched for CI
time_sleep = time.sleep

skipif_emscripten = mark.skipif(
    sys.platform == "emscripten",
    reason="threads, processes, etc. are not available in the browser",
)

no_shrink = tuple(set(settings.default.phases) - {Phase.shrink, Phase.explain})


def flaky(max_runs, min_passes):
    assert isinstance(max_runs, int)
    assert isinstance(min_passes, int)
    assert 0 < min_passes <= max_runs <= 50  # arbitrary cap

    def accept(func):
        @proxies(func)
        def inner(*args, **kwargs):
            runs = passes = 0
            while passes < min_passes:
                runs += 1
                try:
                    func(*args, **kwargs)
                    passes += 1
                except BaseException:
                    if runs >= max_runs:
                        raise

        return inner

    return accept


capture_out_lock = Lock()


@contextlib.contextmanager
def capture_out():
    # replacing the singleton sys.stdout can't be made thread safe. Disallow
    # concurrency by wrapping a lock around the entire block
    with capture_out_lock:
        old_out = sys.stdout
        try:
            new_out = StringIO()
            sys.stdout = new_out
            with with_reporter(default):
                yield new_out
        finally:
            sys.stdout = old_out


class ExcInfo:
    pass


def fails_with(e, *, match=None):
    def accepts(f):
        @proxies(f)
        def inverted_test(*arguments, **kwargs):
            with pytest.raises(e, match=match):
                f(*arguments, **kwargs)

        return inverted_test

    return accepts


fails = fails_with(AssertionError)


class NotDeprecated(Exception):
    pass


@contextlib.contextmanager
def validate_deprecation():

    if settings.get_current_profile_name() == "threading":
        import pytest

        if sys.version_info[:2] < (3, 14):
            pytest.skip("warnings module is not thread-safe before 3.14")

    import warnings

    try:
        warnings.simplefilter("always", HypothesisDeprecationWarning)
        with warnings.catch_warnings(record=True) as w:
            yield
    finally:
        warnings.simplefilter("error", HypothesisDeprecationWarning)
        if not any(e.category == HypothesisDeprecationWarning for e in w):
            raise NotDeprecated(
                f"Expected a deprecation warning but got {[e.category for e in w]!r}"
            )


def checks_deprecated_behaviour(func):
    """A decorator for testing deprecated behaviour."""

    @proxies(func)
    def _inner(*args, **kwargs):
        with validate_deprecation():
            return func(*args, **kwargs)

    return _inner


def all_values(db):
    return {v for vs in db.data.values() for v in vs}


def non_covering_examples(database):
    return {
        v for k, vs in database.data.items() if not k.endswith(b".pareto") for v in vs
    }


def counts_calls(func):
    """A decorator that counts how many times a function was called, and
    stores that value in a ``.calls`` attribute.
    """
    assert not hasattr(func, "calls")

    @proxies(func)
    def _inner(*args, **kwargs):
        _inner.calls += 1
        return func(*args, **kwargs)

    _inner.calls = 0
    return _inner


def assert_output_contains_failure(output, test, **kwargs):
    assert test.__name__ + "(" in output
    for k, v in kwargs.items():
        assert f"{k}={v!r}" in output, (f"{k}={v!r}", output)


def assert_falsifying_output(
    test, example_type="Falsifying", expected_exception=AssertionError, **kwargs
):
    with capture_out() as out:
        if expected_exception is None:
            # Some tests want to check the output of non-failing runs.
            test()
            msg = ""
        else:
            with pytest.raises(expected_exception) as exc_info:
                test()
            notes = "\n".join(getattr(exc_info.value, "__notes__", []))
            msg = str(exc_info.value) + "\n" + notes

    output = out.getvalue() + msg
    assert f"{example_type} example:" in output
    assert_output_contains_failure(output, test, **kwargs)


temp_registered_lock = RLock()


@contextlib.contextmanager
def temp_registered(type_, strat_or_factory):
    """Register and un-register a type for st.from_type().

    This is not too hard, but there's a subtlety in restoring the
    previously-registered strategy which we got wrong in a few places.
    """
    with temp_registered_lock:
        prev = _global_type_lookup.get(type_)
        register_type_strategy(type_, strat_or_factory)
        try:
            yield
        finally:
            del _global_type_lookup[type_]
            from_type.__clear_cache()
            if prev is not None:
                register_type_strategy(type_, prev)


@contextlib.contextmanager
def raises_warning(expected_warning, match=None):
    """Use instead of pytest.warns to check that the raised warning is handled properly"""
    with pytest.raises(expected_warning, match=match) as r, warnings.catch_warnings():
        warnings.simplefilter("error", category=expected_warning)
        yield r


@contextlib.contextmanager
def capture_observations(*, choices=None):
    ls: list[Observation] = []
    add_observability_callback(ls.append)
    if choices is not None:
        old_choices = observability.OBSERVABILITY_CHOICES
        observability.OBSERVABILITY_CHOICES = choices

    try:
        yield ls
    finally:
        remove_observability_callback(ls.append)
        if choices is not None:
            observability.OBSERVABILITY_CHOICES = old_choices


# Specifies whether we can represent subnormal floating point numbers.
# IEE-754 requires subnormal support, but it's often disabled anyway by unsafe
# compiler options like `-ffast-math`.  On most hardware that's even a global
# config option, so *linking against* something built this way can break us.
# Everything is terrible
PYTHON_FTZ = next_down(sys.float_info.min) == 0.0


class Why(enum.Enum):
    # Categorizing known failures, to ease later follow-up investigation.
    # Some are crosshair issues, some hypothesis issues, others truly ok-to-xfail tests.
    symbolic_outside_context = "CrosshairInternal error (using value outside context)"
    nested_given = "nested @given decorators don't work with crosshair"
    undiscovered = "crosshair may not find the failing input"
    other = "reasons not elsewhere categorized"


def xfail_on_crosshair(why: Why, /, *, strict=True, as_marks=False):
    # run `pytest -m xf_crosshair` to select these tests!
    mark = pytest.mark.xfail(
        strict=strict and why != Why.undiscovered,
        reason=f"Expected failure due to: {why.value}",
        condition=settings().backend == "crosshair",
    )
    if as_marks:  # for use with pytest.param(..., marks=xfail_on_crosshair())
        return (pytest.mark.xf_crosshair, mark)
    return lambda fn: pytest.mark.xf_crosshair(mark(fn))


def skipif_threading(f):
    return pytest.mark.skipif(
        settings.get_current_profile_name() == "threading", reason="not thread safe"
    )(f)


def xfail_if_gil_disabled(f):
    try:
        if not sys._is_gil_enabled():  # 3.13+
            return pytest.mark.xfail(
                reason="fails on free-threading build", strict=False
            )(f)
    except Exception:
        pass
    return f


# we don't monkeypatch _consistently_increment_time under threading
skipif_time_unpatched = skipif_threading


_restore_recursion_limit_lock = RLock()


@contextlib.contextmanager
def restore_recursion_limit():
    with _restore_recursion_limit_lock:
        original_limit = sys.getrecursionlimit()
        try:
            yield
        finally:
            sys.setrecursionlimit(original_limit)


def run_concurrently(function, *, n: int) -> None:
    import pytest

    if settings.get_current_profile_name() == "crosshair":
        pytest.skip("crosshair is not thread safe")
    if sys.platform == "emscripten":
        pytest.skip("no threads on emscripten")

    def run():
        barrier.wait()
        function()

    threads = [Thread(target=run) for _ in range(n)]
    barrier = Barrier(n)

    for thread in threads:
        thread.start()

    for thread in threads:
        thread.join(timeout=10)


def wait_for(condition, *, timeout=1, interval=0.01):
    for _ in range(math.ceil(timeout / interval)):
        if condition():
            return
        time_sleep(interval)
    raise Exception(
        f"timing out after waiting {timeout}s for condition "
        f"{get_pretty_function_description(condition)}"
    )


def run_test_for_falsifying_example(test_fn):
    with pytest.raises(AssertionError) as err:
        test_fn()
    return "\n".join(err.value.__notes__).strip()



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import math
import sys
from copy import deepcopy

import pytest

from hypothesis import (
    HealthCheck,
    assume,
    example,
    given,
    note,
    settings,
    strategies as st,
)
from hypothesis.errors import StopTest
from hypothesis.internal.conjecture.choice import (
    ChoiceNode,
    ChoiceTemplate,
    choice_equal,
    choice_from_index,
    choice_permitted,
    choice_to_index,
    choices_key,
)
from hypothesis.internal.conjecture.data import (
    COLLECTION_DEFAULT_MAX_SIZE,
    ConjectureData,
    Status,
    choices_size,
)
from hypothesis.internal.conjecture.datatree import (
    MAX_CHILDREN_EFFECTIVELY_INFINITE,
    all_children,
    compute_max_children,
)
from hypothesis.internal.conjecture.engine import choice_count
from hypothesis.internal.conjecture.provider_conformance import integer_constraints
from hypothesis.internal.floats import SMALLEST_SUBNORMAL, next_down, next_up
from hypothesis.internal.intervalsets import IntervalSet

from tests.common.debug import minimal
from tests.conjecture.common import (
    choice_types_constraints,
    clamped_shrink_towards,
    draw_value,
    float_constr,
    fresh_data,
    integer_constr,
    nodes,
)


# we max out at 128 bit integers in the *unbounded* case, but someone may
# specify a bound with a larger magnitude. Ensure we calculate max children for
# those cases correctly.
@example(("integer", integer_constr(max_value=-(2**200))))
@example(("integer", integer_constr(min_value=2**200)))
@example(("integer", integer_constr(-(2**200), 2**200)))
@given(choice_types_constraints())
def test_compute_max_children_is_positive(choice_type_and_constraints):
    choice_type, constraints = choice_type_and_constraints
    assert compute_max_children(choice_type, constraints) >= 0


@pytest.mark.parametrize(
    "choice_type, constraints, count_children",
    [
        ("integer", {"min_value": 1, "max_value": 2, "weights": {1: 0.1, 2: 0.1}}, 2),
        # only possibility is the empty string
        (
            "string",
            {"min_size": 0, "max_size": 100, "intervals": IntervalSet.from_string("")},
            1,
        ),
        (
            "string",
            {"min_size": 0, "max_size": 0, "intervals": IntervalSet.from_string("abc")},
            1,
        ),
        # 3 possibilities for each character, 8 characters, 3 ** 8 possibilities.
        (
            "string",
            {"min_size": 8, "max_size": 8, "intervals": IntervalSet.from_string("abc")},
            3**8,
        ),
        (
            "string",
            {
                "min_size": 2,
                "max_size": 8,
                "intervals": IntervalSet.from_string("abcd"),
            },
            sum(4**k for k in range(2, 8 + 1)),
        ),
        (
            "string",
            {
                "min_size": 0,
                "max_size": COLLECTION_DEFAULT_MAX_SIZE,
                "intervals": IntervalSet.from_string("a"),
            },
            COLLECTION_DEFAULT_MAX_SIZE + 1,
        ),
        (
            "string",
            {
                "min_size": 0,
                "max_size": 10_000,
                "intervals": IntervalSet.from_string("abcdefg"),
            },
            MAX_CHILDREN_EFFECTIVELY_INFINITE,
        ),
        (
            "bytes",
            {
                "min_size": 0,
                "max_size": 2,
            },
            sum(2 ** (8 * k) for k in range(2 + 1)),
        ),
        (
            "bytes",
            {
                "min_size": 0,
                "max_size": COLLECTION_DEFAULT_MAX_SIZE,
            },
            MAX_CHILDREN_EFFECTIVELY_INFINITE,
        ),
        (
            "bytes",
            {
                "min_size": 0,
                "max_size": 10_000,
            },
            MAX_CHILDREN_EFFECTIVELY_INFINITE,
        ),
        ("boolean", {"p": 0.0}, 1),
        ("boolean", {"p": 1.0}, 1),
        ("boolean", {"p": 0.5}, 2),
        ("boolean", {"p": 0.001}, 2),
        ("boolean", {"p": 0.999}, 2),
        ("float", float_constr(0.0, 0.0), 1),
        ("float", float_constr(-0.0, -0.0), 1),
        ("float", float_constr(-0.0, 0.0), 2),
        ("float", float_constr(next_down(-0.0), next_up(0.0)), 4),
        (
            "float",
            float_constr(
                next_down(next_down(-0.0)),
                next_up(next_up(0.0)),
                smallest_nonzero_magnitude=next_up(SMALLEST_SUBNORMAL),
            ),
            4,
        ),
        ("float", float_constr(smallest_nonzero_magnitude=next_down(math.inf)), 6),
        ("float", float_constr(1, 10, smallest_nonzero_magnitude=11.0), 0),
        ("float", float_constr(-3, -2, smallest_nonzero_magnitude=4.0), 0),
    ],
)
def test_compute_max_children(choice_type, constraints, count_children):
    assert compute_max_children(choice_type, constraints) == count_children


@given(st.text(min_size=1, max_size=1), st.integers(0, 100))
def test_draw_string_single_interval_with_equal_bounds(s, n):
    data = fresh_data()
    intervals = IntervalSet.from_string(s)
    assert data.draw_string(intervals, min_size=n, max_size=n) == s * n


@example(("boolean", {"p": 2**-65}))
@example(("boolean", {"p": 1 - 2**-65}))
@example(
    (
        "string",
        {"min_size": 0, "max_size": 0, "intervals": IntervalSet.from_string("abc")},
    )
)
@example(
    ("string", {"min_size": 0, "max_size": 3, "intervals": IntervalSet.from_string("")})
)
@example(
    (
        "string",
        {"min_size": 0, "max_size": 3, "intervals": IntervalSet.from_string("a")},
    )
)
# all combinations of float signs
@example(("float", float_constr(next_down(-0.0), -0.0)))
@example(("float", float_constr(next_down(-0.0), next_up(0.0))))
@example(("float", float_constr(0.0, next_up(0.0))))
# using a smallest_nonzero_magnitude which happens to filter out everything
@example(("float", float_constr(1.0, 2.0, smallest_nonzero_magnitude=3.0)))
@example(("integer", integer_constr(1, 2, weights={1: 0.2, 2: 0.4})))
@given(choice_types_constraints())
@settings(suppress_health_check=[HealthCheck.filter_too_much])
def test_compute_max_children_and_all_children_agree(choice_type_and_constraints):
    choice_type, constraints = choice_type_and_constraints
    max_children = compute_max_children(choice_type, constraints)

    # avoid slowdowns / OOM when reifying extremely large all_children generators.
    # We also hard cap at MAX_CHILDREN_EFFECTIVELY_INFINITE, because max_children
    # returns approximations after this value and so will disagree with
    # all_children.
    cap = min(100_000, MAX_CHILDREN_EFFECTIVELY_INFINITE)
    assume(max_children < cap)
    assert len(list(all_children(choice_type, constraints))) == max_children


# it's very hard to test that unbounded integer ranges agree with
# compute_max_children, because they by necessity require iterating over 2**127
# or more elements. We do the not great approximation of checking just the first
# element is what we expect.


@given(integer_constraints())
def test_compute_max_children_unbounded_integer_ranges(constraints):
    expected = clamped_shrink_towards(constraints)
    first = next(all_children("integer", constraints))
    assert expected == first, (expected, first)


@given(st.randoms())
def test_nodes(random):
    data = fresh_data(random=random)
    data.draw_float(min_value=-10.0, max_value=10.0, forced=5.0)
    data.draw_boolean(forced=True)

    data.start_span(42)
    data.draw_string(IntervalSet.from_string("abcd"), forced="abbcccdddd")
    data.draw_bytes(8, 8, forced=bytes(8))
    data.stop_span()

    data.draw_integer(0, 100, forced=50)

    data.freeze()
    expected_tree_nodes = (
        ChoiceNode(
            type="float",
            value=5.0,
            constraints=float_constr(-10.0, 10.0),
            was_forced=True,
        ),
        ChoiceNode(
            type="boolean",
            value=True,
            constraints={"p": 0.5},
            was_forced=True,
        ),
        ChoiceNode(
            type="string",
            value="abbcccdddd",
            constraints={
                "intervals": IntervalSet.from_string("abcd"),
                "min_size": 0,
                "max_size": COLLECTION_DEFAULT_MAX_SIZE,
            },
            was_forced=True,
        ),
        ChoiceNode(
            type="bytes",
            value=bytes(8),
            constraints={"min_size": 8, "max_size": 8},
            was_forced=True,
        ),
        ChoiceNode(
            type="integer",
            value=50,
            constraints=integer_constr(0, 100),
            was_forced=True,
        ),
    )
    assert data.nodes == expected_tree_nodes


@given(nodes())
def test_copy_choice_node(node):
    assert node == node

    assume(not node.was_forced)
    new_value = draw_value(node.type, node.constraints)
    # if we drew the same value as before, the node should still be equal
    assert (node.copy(with_value=new_value) == node) is (
        choice_equal(new_value, node.value)
    )


@given(nodes())
def test_choice_node_equality(node):
    assert node == node
    # for coverage on our NotImplemented return, more than anything.
    assert node != 42


@given(nodes(was_forced=True))
def test_cannot_modify_forced_nodes(node):
    with pytest.raises(AssertionError):
        node.copy(with_value=42)


def test_data_with_empty_choices_is_overrun():
    data = ConjectureData.for_choices([])
    with pytest.raises(StopTest):
        data.draw_integer()

    assert data.status is Status.OVERRUN


@given(nodes(was_forced=True))
def test_data_with_changed_forced_value(node):
    # we had a forced node and then tried to draw a different forced value from it.
    # nodes:   v1 [was_forced=True]
    # drawing:    [forced=v2]
    #
    # This is actually fine; we'll just ignore the forced node (v1) and return
    # what the draw expects (v2).

    data = ConjectureData.for_choices([node.value])

    draw_func = getattr(data, f"draw_{node.type}")
    constraints = deepcopy(node.constraints)
    constraints["forced"] = draw_value(node.type, node.constraints)
    assume(not choice_equal(constraints["forced"], node.value))

    assert choice_equal(draw_func(**constraints), constraints["forced"])


# ensure we hit bare-minimum coverage for all choice sequence types.
@example(
    ChoiceNode(type="float", value=0.0, constraints=float_constr(), was_forced=True)
)
@example(
    ChoiceNode(
        type="boolean",
        value=False,
        constraints={"p": 0.5},
        was_forced=True,
    )
)
@example(
    ChoiceNode(
        type="integer", value=50, constraints=integer_constr(50, 100), was_forced=True
    )
)
@example(
    ChoiceNode(
        type="string",
        value="aaaa",
        constraints={
            "intervals": IntervalSet.from_string("bcda"),
            "min_size": 4,
            "max_size": COLLECTION_DEFAULT_MAX_SIZE,
        },
        was_forced=True,
    )
)
@example(
    ChoiceNode(
        type="bytes",
        value=bytes(8),
        constraints={"min_size": 8, "max_size": 8},
        was_forced=True,
    )
)
@given(nodes(was_forced=True))
def test_data_with_same_forced_value_is_valid(node):
    # we had a forced node and then drew the same forced value. This is totally
    # fine!
    # nodes:   v1 [was_forced=True]
    # drawing:    [forced=v1]
    data = ConjectureData.for_choices([node.value])
    draw_func = getattr(data, f"draw_{node.type}")

    constraints = deepcopy(node.constraints)
    constraints["forced"] = node.value
    assert choice_equal(draw_func(**constraints), constraints["forced"])


@given(choice_types_constraints())
@settings(suppress_health_check=[HealthCheck.filter_too_much])
def test_all_children_are_permitted_values(choice_type_and_constraints):
    choice_type, constraints = choice_type_and_constraints
    max_children = compute_max_children(choice_type, constraints)

    cap = min(100_000, MAX_CHILDREN_EFFECTIVELY_INFINITE)
    assume(max_children < cap)

    # test that all_children -> choice_permitted (but not necessarily the converse.)
    for value in all_children(choice_type, constraints):
        assert choice_permitted(value, constraints), value


@pytest.mark.parametrize(
    "value, constraints, permitted",
    [
        (0, integer_constr(1, 2), False),
        (2, integer_constr(0, 1), False),
        (10, integer_constr(0, 20), True),
        (int(2**128 / 2) - 1, integer_constr(), True),
        (int(2**128 / 2), integer_constr(), True),
        (math.nan, float_constr(0.0, 0.0), True),
        (math.nan, float_constr(0.0, 0.0, allow_nan=False), False),
        (2.0, float_constr(1.0, 3.0, smallest_nonzero_magnitude=2.5), False),
        (
            -2.0,
            float_constr(-3.0, -1.0, smallest_nonzero_magnitude=2.5),
            False,
        ),
        (1.0, float_constr(1.0, 1.0), True),
        (
            "abcd",
            {
                "min_size": 10,
                "max_size": 20,
                "intervals": IntervalSet.from_string("abcd"),
            },
            False,
        ),
        (
            "abcd",
            {
                "min_size": 1,
                "max_size": 3,
                "intervals": IntervalSet.from_string("abcd"),
            },
            False,
        ),
        (
            "abcd",
            {"min_size": 1, "max_size": 10, "intervals": IntervalSet.from_string("e")},
            False,
        ),
        (
            "e",
            {"min_size": 1, "max_size": 10, "intervals": IntervalSet.from_string("e")},
            True,
        ),
        (b"a", {"min_size": 2, "max_size": 2}, False),
        (b"aa", {"min_size": 2, "max_size": 2}, True),
        (b"aa", {"min_size": 0, "max_size": 3}, True),
        (b"a", {"min_size": 2, "max_size": 10}, False),
        (True, {"p": 0}, False),
        (False, {"p": 0}, True),
        (True, {"p": 1}, True),
        (False, {"p": 1}, False),
        (True, {"p": 0.5}, True),
        (False, {"p": 0.5}, True),
    ],
)
def test_choice_permitted(value, constraints, permitted):
    assert choice_permitted(value, constraints) == permitted


@given(nodes(was_forced=True))
def test_forced_nodes_are_trivial(node):
    assert node.trivial


@pytest.mark.parametrize(
    "node",
    [
        ChoiceNode(
            type="float",
            value=5.0,
            constraints=float_constr(5.0, 10.0),
            was_forced=False,
        ),
        ChoiceNode(
            type="float",
            value=0.0,
            constraints=float_constr(-5.0, 5.0),
            was_forced=False,
        ),
        ChoiceNode(
            type="float", value=0.0, constraints=float_constr(), was_forced=False
        ),
        ChoiceNode(
            type="boolean", value=False, constraints={"p": 0.5}, was_forced=False
        ),
        ChoiceNode(
            type="boolean", value=True, constraints={"p": 1.0}, was_forced=False
        ),
        ChoiceNode(
            type="boolean", value=False, constraints={"p": 0.0}, was_forced=False
        ),
        ChoiceNode(
            type="string",
            value="",
            constraints={
                "intervals": IntervalSet.from_string("abcd"),
                "min_size": 0,
                "max_size": COLLECTION_DEFAULT_MAX_SIZE,
            },
            was_forced=False,
        ),
        ChoiceNode(
            type="string",
            value="aaaa",
            constraints={
                "intervals": IntervalSet.from_string("bcda"),
                "min_size": 4,
                "max_size": COLLECTION_DEFAULT_MAX_SIZE,
            },
            was_forced=False,
        ),
        ChoiceNode(
            type="bytes",
            value=bytes(8),
            constraints={"min_size": 8, "max_size": 8},
            was_forced=False,
        ),
        ChoiceNode(
            type="bytes",
            value=bytes(2),
            constraints={"min_size": 2, "max_size": COLLECTION_DEFAULT_MAX_SIZE},
            was_forced=False,
        ),
        ChoiceNode(
            type="integer",
            value=50,
            constraints=integer_constr(50, 100),
            was_forced=False,
        ),
        ChoiceNode(
            type="integer",
            value=0,
            constraints=integer_constr(-10, 10),
            was_forced=False,
        ),
        ChoiceNode(
            type="integer",
            value=2,
            constraints=integer_constr(-10, 10, shrink_towards=2),
            was_forced=False,
        ),
        ChoiceNode(
            type="integer",
            value=-10,
            constraints=integer_constr(-10, 10, shrink_towards=-12),
            was_forced=False,
        ),
        ChoiceNode(
            type="integer",
            value=10,
            constraints=integer_constr(-10, 10, shrink_towards=12),
            was_forced=False,
        ),
        ChoiceNode(
            type="integer", value=0, constraints=integer_constr(), was_forced=False
        ),
        ChoiceNode(
            type="integer",
            value=1,
            constraints=integer_constr(min_value=-10, shrink_towards=1),
            was_forced=False,
        ),
        ChoiceNode(
            type="integer",
            value=1,
            constraints=integer_constr(max_value=10, shrink_towards=1),
            was_forced=False,
        ),
        ChoiceNode(
            type="integer",
            value=1,
            constraints={
                "min_value": None,
                "max_value": None,
                "weights": None,
                "shrink_towards": 1,
            },
            was_forced=False,
        ),
    ],
)
def test_trivial_nodes(node):
    assert node.trivial

    @st.composite
    def values(draw):
        data = draw(st.data()).conjecture_data
        return getattr(data, f"draw_{node.type}")(**node.constraints)

    # if we're trivial, then shrinking should produce the same value.
    assert choice_equal(minimal(values()), node.value)


@pytest.mark.parametrize(
    "node",
    [
        ChoiceNode(
            type="float",
            value=6.0,
            constraints=float_constr(5.0, 10.0),
            was_forced=False,
        ),
        ChoiceNode(
            type="float",
            value=-5.0,
            constraints=float_constr(-5.0, 5.0),
            was_forced=False,
        ),
        ChoiceNode(
            type="float", value=1.0, constraints=float_constr(), was_forced=False
        ),
        ChoiceNode(
            type="boolean", value=True, constraints={"p": 0.5}, was_forced=False
        ),
        ChoiceNode(
            type="boolean", value=True, constraints={"p": 0.99}, was_forced=False
        ),
        ChoiceNode(
            type="string",
            value="d",
            constraints={
                "intervals": IntervalSet.from_string("abcd"),
                "min_size": 1,
                "max_size": COLLECTION_DEFAULT_MAX_SIZE,
            },
            was_forced=False,
        ),
        ChoiceNode(
            type="bytes",
            value=b"\x01",
            constraints={"min_size": 1, "max_size": 1},
            was_forced=False,
        ),
        ChoiceNode(
            type="bytes",
            value=bytes(1),
            constraints={"min_size": 0, "max_size": COLLECTION_DEFAULT_MAX_SIZE},
            was_forced=False,
        ),
        ChoiceNode(
            type="bytes",
            value=bytes(2),
            constraints={"min_size": 1, "max_size": 10},
            was_forced=False,
        ),
        ChoiceNode(
            type="integer",
            value=-10,
            constraints=integer_constr(-10, 10),
            was_forced=False,
        ),
        ChoiceNode(
            type="integer", value=42, constraints=integer_constr(), was_forced=False
        ),
    ],
)
def test_nontrivial_nodes(node):
    assert not node.trivial

    @st.composite
    def values(draw):
        data = draw(st.data()).conjecture_data
        return getattr(data, f"draw_{node.type}")(**node.constraints)

    # if we're nontrivial, then shrinking should produce something different.
    assert not choice_equal(minimal(values()), node.value)


@pytest.mark.parametrize(
    "node",
    [
        ChoiceNode(
            type="float",
            value=1.5,
            constraints=float_constr(1.1, 1.6),
            was_forced=False,
        ),
        ChoiceNode(
            type="float",
            value=float(math.floor(sys.float_info.max)),
            constraints=float_constr(sys.float_info.max - 1, math.inf),
            was_forced=False,
        ),
        ChoiceNode(
            type="float",
            value=float(math.ceil(-sys.float_info.max)),
            constraints=float_constr(-math.inf, -sys.float_info.max + 1),
            was_forced=False,
        ),
        ChoiceNode(
            type="float",
            value=math.inf,
            constraints=float_constr(math.inf, math.inf),
            was_forced=False,
        ),
        ChoiceNode(
            type="float",
            value=-math.inf,
            constraints=float_constr(-math.inf, -math.inf),
            was_forced=False,
        ),
    ],
)
def test_conservative_nontrivial_nodes(node):
    # these nodes actually are trivial, but our analysis doesn't compute them
    # as such. We'd like to improve this in the future!
    assert not node.trivial

    @st.composite
    def values(draw):
        data = draw(st.data()).conjecture_data
        return getattr(data, f"draw_{node.type}")(**node.constraints)

    assert choice_equal(minimal(values()), node.value)


@given(nodes())
def test_choice_node_is_hashable(node):
    hash(node)


@given(st.lists(nodes()))
def test_choices_size_positive(nodes):
    assert choices_size([n.value for n in nodes]) >= 0


@given(st.integers(min_value=1))
def test_node_template_count(n):
    node = ChoiceTemplate(type="simplest", count=n)
    assert choice_count([node]) == n


def test_node_template_to_overrun():
    data = ConjectureData.for_choices([1, ChoiceTemplate("simplest", count=5)])
    data.draw_integer()
    with pytest.raises(StopTest):
        for _ in range(10):
            data.draw_integer()

    assert data.status is Status.OVERRUN


def test_node_template_single_node_overruns():
    # test for when drawing a single node takes more than BUFFER_SIZE, while in
    # the ChoiceTemplate case
    data = ConjectureData.for_choices((ChoiceTemplate("simplest", count=1),))
    with pytest.raises(StopTest):
        data.draw_bytes(10_000, 10_000)

    assert data.status is Status.OVERRUN


@given(nodes())
def test_node_template_simplest_is_actually_trivial(node):
    # TODO_IR node.trivial is sound but not complete for floats.
    assume(node.type != "float")
    data = ConjectureData.for_choices((ChoiceTemplate("simplest", count=1),))
    getattr(data, f"draw_{node.type}")(**node.constraints)
    assert len(data.nodes) == 1
    assert data.nodes[0].trivial


@given(choice_types_constraints())
@example(("boolean", {"p": 0}))
@example(("boolean", {"p": 1}))
def test_choice_indices_are_positive(choice_type_and_constraints):
    choice_type, constraints = choice_type_and_constraints
    v = draw_value(choice_type, constraints)
    assert choice_to_index(v, constraints) >= 0


@given(integer_constraints())
def test_shrink_towards_has_index_0(constraints):
    shrink_towards = clamped_shrink_towards(constraints)
    note({"clamped_shrink_towards": shrink_towards})
    assert choice_to_index(shrink_towards, constraints) == 0
    assert choice_from_index(0, "integer", constraints) == shrink_towards


@given(choice_types_constraints())
@settings(max_examples=20)
def test_choice_to_index_injective(choice_type_and_constraints):
    # choice sequence ordering should be injective both ways.
    choice_type, constraints = choice_type_and_constraints
    # ...except for floats, which are hard to order bijectively.
    assume(choice_type != "float")
    # cap to 10k so this test finishes in a reasonable amount of time
    cap = min(compute_max_children(choice_type, constraints), 10_000)

    indices = set()
    for i, choice in enumerate(all_children(choice_type, constraints)):
        if i >= cap:
            break
        index = choice_to_index(choice, constraints)
        assert index not in indices
        indices.add(index)


@given(choice_types_constraints())
@settings(max_examples=10)
@example(
    (
        "string",
        {"min_size": 0, "max_size": 10, "intervals": IntervalSet.from_string("a")},
    )
)
def test_choice_from_value_injective(choice_type_and_constraints):
    choice_type, constraints = choice_type_and_constraints
    assume(choice_type != "float")
    cap = min(compute_max_children(choice_type, constraints), 10_000)

    choices = set()
    for index in range(cap):
        choice = choice_from_index(index, choice_type, constraints)
        assert choice not in choices
        choices.add(choice)


@given(choice_types_constraints())
def test_choice_index_and_value_are_inverses(choice_type_and_constraints):
    choice_type, constraints = choice_type_and_constraints
    v = draw_value(choice_type, constraints)
    index = choice_to_index(v, constraints)
    note({"v": v, "index": index})
    choice_equal(choice_from_index(index, choice_type, constraints), v)


@pytest.mark.parametrize(
    "choice_type, constraints, choices",
    [
        ("boolean", {"p": 1}, [True]),
        ("boolean", {"p": 0}, [False]),
        ("integer", integer_constr(min_value=1, shrink_towards=4), range(1, 10)),
        ("integer", integer_constr(max_value=5, shrink_towards=2), range(-10, 5 + 1)),
        ("integer", integer_constr(max_value=5), range(-10, 5 + 1)),
        ("integer", integer_constr(min_value=0, shrink_towards=1), range(10)),
        ("integer", integer_constr(-5, 5, shrink_towards=3), range(-5, 5 + 1)),
        ("integer", integer_constr(-5, 5, shrink_towards=-3), range(-5, 5 + 1)),
        (
            "float",
            float_constr(1.0, next_up(next_up(1.0))),
            [1.0, next_up(1.0), next_up(next_up(1.0))],
        ),
        (
            "float",
            float_constr(next_down(-0.0), next_up(0.0)),
            [next_down(-0.0), -0.0, 0.0, next_up(0.0)],
        ),
    ],
)
def test_choice_index_and_value_are_inverses_explicit(
    choice_type, constraints, choices
):
    for choice in choices:
        index = choice_to_index(choice, constraints)
        assert choice_equal(choice_from_index(index, choice_type, constraints), choice)


@pytest.mark.parametrize(
    "constraints, choices",
    [
        # unbounded
        (integer_constr(), (0, 1, -1, 2, -2, 3, -3)),
        (integer_constr(shrink_towards=2), (2, 3, 1, 4, 0, 5, -1, 6, -2)),
        # semibounded (below)
        (integer_constr(min_value=3), (3, 4, 5, 6, 7)),
        (integer_constr(min_value=3, shrink_towards=5), (5, 6, 4, 7, 3, 8, 9)),
        (integer_constr(min_value=-3), (0, 1, -1, 2, -2, 3, -3, 4, 5, 6)),
        (integer_constr(min_value=-3, shrink_towards=-1), (-1, 0, -2, 1, -3, 2, 3, 4)),
        # semibounded (above)
        (integer_constr(max_value=3), (0, 1, -1, 2, -2, 3, -3, -4, -5, -6)),
        (integer_constr(max_value=3, shrink_towards=1), (1, 2, 0, 3, -1, -2, -3, -4)),
        (integer_constr(max_value=-3), (-3, -4, -5, -6, -7)),
        (integer_constr(max_value=-3, shrink_towards=-5), (-5, -4, -6, -3, -7, -8, -9)),
        # bounded
        (integer_constr(-3, 3), (0, 1, -1, 2, -2, 3, -3)),
        (integer_constr(-3, 3, shrink_towards=1), (1, 2, 0, 3, -1, -2, -3)),
        (integer_constr(-3, 3, shrink_towards=-1), (-1, 0, -2, 1, -3, 2, 3)),
    ],
    ids=repr,
)
def test_integer_choice_index(constraints, choices):
    # explicit test which checks that the order of `choices` matches the index
    # order.
    for i, choice in enumerate(choices):
        assert choice_to_index(choice, constraints) == i


@given(st.lists(nodes()))
def test_drawing_directly_matches_for_choices(nodes):
    data = ConjectureData.for_choices([n.value for n in nodes])
    for node in nodes:
        value = getattr(data, f"draw_{node.type}")(**node.constraints)
        assert choice_equal(node.value, value)


def test_draw_directly_explicit():
    # this is a much weaker and more explicit variant of the property-based test
    # directly above, but this is such an important thing to ensure that we have
    # correct that it's worth some duplication in case we ever screw up our pbt test.
    assert (
        ConjectureData.for_choices(["a"]).draw_string(
            IntervalSet([(0, 127)]), min_size=1
        )
        == "a"
    )
    assert ConjectureData.for_choices([b"a"]).draw_bytes() == b"a"
    assert (
        ConjectureData.for_choices([1.0]).draw_float(
            0.0, 2.0, allow_nan=False, smallest_nonzero_magnitude=0.5
        )
        == 1.0
    )
    assert ConjectureData.for_choices([True]).draw_boolean(0.3)
    assert ConjectureData.for_choices([42]).draw_integer() == 42
    assert (
        ConjectureData.for_choices([-42]).draw_integer(min_value=-50, max_value=0)
        == -42
    )
    assert (
        ConjectureData.for_choices([10]).draw_integer(
            min_value=10, max_value=11, weights={10: 0.1, 11: 0.3}
        )
        == 10
    )


@pytest.mark.parametrize(
    "choices1, choices2",
    [
        [(True,), (1,)],
        [(True,), (1.0,)],
        [(False,), (0,)],
        [(False,), (0.0,)],
        [(False,), (-0.0,)],
        [(0.0,), (-0.0,)],
    ],
)
def test_choices_key_distinguishes_weird_cases(choices1, choices2):
    assert choices_key(choices1) != choices_key(choices2)


def test_node_template_overrun():
    # different code path for overruning the ChoiceTemplate count, not BUFFER_SIZE.
    cd = ConjectureData(
        random=None,
        prefix=[ChoiceTemplate("simplest", count=2)],
        max_choices=100,
    )

    cd.draw_integer()
    cd.draw_integer()
    try:
        cd.draw_integer()
    except StopTest:
        pass
    assert cd.status is Status.OVERRUN



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Minimal setup.py to register an entrypoint."""

import setuptools

setuptools.setup(
    name="example_hypothesis_entrypoint",
    author="Example Author",
    email="author@example.com",
    license="MPL v2",
    description="Minimal setup.py to register an entrypoint.",
    packages=setuptools.find_packages(),
    install_requires=["hypothesis"],
    python_requires=">=3.7",
    entry_points={
        "hypothesis": ["_ = example_hypothesis_entrypoint:_hypothesis_setup_hook"]
    },
)



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

import io
import sys
import unittest

import pytest

from hypothesis import given, strategies as st
from hypothesis.errors import FailedHealthCheck, HypothesisWarning

from tests.common.utils import fails_with, skipif_emscripten


class Thing_with_a_subThing(unittest.TestCase):
    """Example test case using subTest for the actual test below."""

    @given(st.tuples(st.booleans(), st.booleans()))
    def thing(self, lst):
        for i, b in enumerate(lst):
            with pytest.warns(HypothesisWarning), self.subTest((i, b)):
                self.assertTrue(b)


def test_subTest():
    suite = unittest.TestSuite()
    suite.addTest(Thing_with_a_subThing("thing"))
    stream = io.StringIO()
    out = unittest.TextTestRunner(stream=stream).run(suite)
    assert len(out.failures) <= out.testsRun, out


class test_given_on_setUp_fails_health_check(unittest.TestCase):
    @fails_with(FailedHealthCheck)
    @given(st.integers())
    def setUp(self, i):
        pass

    def test(self):
        """Provide something to set up for, so the setUp method is called."""


SUBTEST_SUITE = """
import unittest
from hypothesis import given, settings, strategies as st

class MyTest(unittest.TestCase):
    @given(s=st.text())
    @settings(deadline=None)
    def test_subtest(self, s):
        with self.subTest(text=s):
            self.assertIsInstance(s, str)

if __name__ == "__main__":
    unittest.main()
"""


@skipif_emscripten
@pytest.mark.parametrize("err", [[], ["-Werror"]])
def test_subTest_no_self(testdir, err):
    # https://github.com/HypothesisWorks/hypothesis/issues/2462
    # for some reason this issue happens only when running unittest from commandline
    fname = testdir.makepyfile(tests=SUBTEST_SUITE)
    result = testdir.run(sys.executable, *err, str(fname))
    expected = pytest.ExitCode.TESTS_FAILED if err else pytest.ExitCode.OK
    assert result.ret == expected, result.stderr.str()



# This file is part of Hypothesis, which may be found at
# https://github.com/HypothesisWorks/hypothesis/
#
# Copyright the Hypothesis Authors.
# Individual contributors are listed in AUTHORS.rst and the git log.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

from hypothesis import given, strategies as st
from hypothesis.strategies._internal.types import _global_type_lookup

from tests.common.debug import find_any

TYPES = sorted(
    (
        x
        for x in _global_type_lookup
        if x.__module__ != "typing" and x.__name__ != "ByteString"
    ),
    key=str,
)


def everything_except(excluded_types):
    """Recipe copied from the docstring of ``from_type``"""
    return (
        st.from_type(type)
        .flatmap(st.from_type)
        .filter(lambda x: not isinstance(x, excluded_types))
    )


@given(
    excluded_types=st.lists(
        st.sampled_from(TYPES), min_size=1, max_size=3, unique=True
    ).map(tuple),
    data=st.data(),
)
def test_recipe_for_everything_except(excluded_types, data):
    value = data.draw(everything_except(excluded_types))
    assert not isinstance(value, excluded_types)


def test_issue_4144_regression():
    find_any(everything_except(()), lambda t: t is not type)
